from .instruction import Instruction

class InstructionFactories:
    def reach_target(
            target: str,
            sources: list=None,
            transformator: str=None,
            forward: dict={}
        ) -> Instruction:

        assert isinstance(target, str)
        assert isinstance(sources, list) or sources is None
        assert isinstance(transformator, str) or transformator is None

        name = f"Reach target '{target}'"

        return Instruction({
            'type': 'transformator_call',
            'name': name,
            'target': target,
            'sources': sources,
            'transformator_name': transformator,
            'forward': forward
        })
    
    def prepare_transform(
            transformator: str,
            transform: str,
            path: list,
            forward: dict={}
        ) -> Instruction:
        
        assert isinstance(transformator, str)
        assert isinstance(transform, str)

        return Instruction({
            'type': 'prepare_transform',
            'name': f"Prepare transform '{transform}' of transformator '{transformator}'",
            'transformator_name': transformator,
            'transform_name': transform,
            'transform_path': path,
            'forward': forward
        })
    
    def execute_transform(
            transformator: str,
            transform: str,
            unmerged_keys: dict,
            forward: dict={}
        ) -> Instruction:
        
        assert isinstance(transformator, str)
        assert isinstance(transform, str)
        
        keys_repr = ', '.join([
            f"{key}={unmerged_keys[key]}"
            for key in sorted(unmerged_keys.keys())
        ])

        return Instruction({
            'type': 'execute_transform',
            'name': f"Execute transform '{transform}' with unmerged keys {keys_repr}",
            'transformator_name': transformator,
            'transform_name': transform,
            'unmerged_keys': unmerged_keys,
            'forward': forward
        })



import os
import copy
import string
import fnmatch
import time
from pathlib import Path

import numpy as np
import pandas as pd
from ..dataitems import REGISTERED_DATAITEM_TYPES, ColumnType, PathItem
from ..utils import get_logger_shortcuts


class DataStorage:
    def __init__(self,
            item_list: list=None,
            logger=None, wd: str=None,
            instantiate_prototypes: bool=True,
            allow_overwrite: bool=False
        ) -> None:

        self.data = {}
        
        self.logger = logger
        self.log = get_logger_shortcuts(logger)
        
        self.additional_kwargs = {}
        if logger is not None:
            self.additional_kwargs['logger'] = logger
        self.additional_kwargs['allow_overwrite'] = allow_overwrite
        
        if wd is None:
            self.wd = os.path.abspath('.')
        else:
            self.wd = os.path.abspath(wd)
        self.log.info(f"Using '{self.wd}' as the base directory for the project")

        self._initialize_directories_item()
        self._initialize_modtimes_item()

        self.items_prototypes = copy.copy(item_list)
        self.instantiate_prototypes = instantiate_prototypes

        if item_list is not None and self.instantiate_prototypes:
            for item_name, item_attrs in item_list.items():
                if 'type' in item_attrs:
                    item_attrs['item_type'] = item_attrs['type']
                    del item_attrs['type']
                self.new_item(name=item_name, **item_attrs)
    
    def new_item(self, name: str, item_type: str, **kwargs) -> None:
        """Initialize a new DataItem to storage

        Args:
            name (str): Name of the new DataItem
            item_type (str): Type of the DataItem as recorded in REGISTERED_DATAITEM_TYPES
            **kwargs: everything that needs to be passed into constructor of the requeted type of DataItem
        """
        assert name not in self.data, f"Item '{name}' is already registered"

        assert item_type in REGISTERED_DATAITEM_TYPES, \
            f"DataItem type '{item_type}' is not registered " \
            f"(registered types are {repr(tuple(REGISTERED_DATAITEM_TYPES.keys()))})"
        
        pass_kwargs = {
            **kwargs,
            **{
                key: value
                for key, value in self.additional_kwargs.items()
                if key not in kwargs
            }
        }
        data_item = REGISTERED_DATAITEM_TYPES[item_type](name=name, storage=self, **pass_kwargs)

        assert hasattr(data_item, 'modtime_control')
        assert isinstance(data_item.modtime_control, dict) and \
                    len(data_item.modtime_control) == 1 and \
                    'column' in data_item.modtime_control or \
                data_item.modtime_control == 'manual'

        self.data[name] = {
            'name': name,
            'type': item_type,
            'item': data_item,
            'table': self._create_table(data_item._key_names),
            'modtime_control': data_item.modtime_control,
            'restrictions': {},
        }
        self.log.info(f"DataItem '{name}' of type '{item_type}' is registered successfully")

        # 1) Create DirectoriesItem that is stored at self.wd
        # 2) Include '.' as the only element
        # 3) FileItems in general, when checking-in first of all do a request for underlying wd to be created
        data_item.checkin()

        return data_item

    def _initialize_directories_item(self) -> None:
        self.new_item(
            name=self._internal('directories'),
            item_type='dirs',
            mask='.',
            wd='.',
            skip_autoindexing=True,
            additional_keys={
                self._internal('itemname'): ColumnType.UNIQUE,
                self._internal('dirobject'): ColumnType.CONSTANT,
            }
        )
        # print(repr(self.data[self._internal('directories')]['table']))

    def _initialize_modtimes_item(self) -> None:
        self.new_item(
            name=self._internal('modtimes'),
            item_type='object',
            keys=[],
            additional_keys={
                self._internal('itemname'): ColumnType.UNIQUE,
                self._internal('restriction_id'): ColumnType.UNIQUE,
                self._internal('element_index'): ColumnType.UNIQUE,
                self._internal('modtime'): ColumnType.VARIABLE,
            }
        )

    def new_restricted_item(self,
            name: str,
            key_restrictions: dict,
            restriction_id: int=None,
            constructor_kwargs={}
        ):
        """Initialize a new restriction of an existing DataItem

        Args:
            name (str): name of the base DataItem
            key_restrictions (dict): key restrictions to be imposed on the new DataItem
            restriction_id (int, optional): restriction ID in cases when
            a new restriction is created not from the main DataItem but from
            another restriction. Defaults to None.
            constructor_kwargs (dict, optional): will be passed to constructor of the new DataItem.
            Defaults to {}.

        Returns:
            DataItem or child: the created restriction
            (note that `checkin` method is not called)
        """

        assert restriction_id is None, 'Not implemented'

        # Get ID for new restriction
        new_restriction_id = 0
        for id in sorted(self.data[name]['restrictions'].keys()):
            if id == new_restriction_id:
                new_restriction_id += 1
            else:
                break
        
        item_type = self.data[name]['type']
        data_item = REGISTERED_DATAITEM_TYPES[item_type](
            name=name,
            storage=self,
            restrictions=copy.copy(key_restrictions),
            restriction_id=new_restriction_id,
            **constructor_kwargs,
        )

        self.data[name]['restrictions'][new_restriction_id] = {
            'name': name,
            'fixed_keys': data_item.restrictions,
            'item': data_item,

            # Keep the original rows that comply with restrictions. TODO Add kwarg to remove fixed columns
            'table': self._request_table(
                name=name,
                key_restrictions=key_restrictions,
                restriction_id=restriction_id,
            ).copy(),
        }
        return data_item

    def _create_table(self, keys: list) -> pd.DataFrame:
        """Prepare empty pandas df that will later be used to
        access/filter/include separate items by their values of keys.

        Args:
            keys (List[str]): key list of the DataItem being created

        Returns:
            pd.DataFrame
        """

        return pd.DataFrame(columns=keys)

    def _access_item_record(self,
            name: str,
            key_restrictions: dict,
            restriction_id: int = None,
            no_additional_restrictions: bool = False
        ) -> dict:
        """This method abstracts access to main record of DataItem and
        calls to access restricted DataItems

        Args:
            name (str): name of the dataitem
            key_restrictions (dict): key restrictions to be applied
            restriction_id (int, optional): ID of restricted item
            no_additional_restrictions (bool): enables additional assertion that no
            restrictions are present (except the base restrictions of restricted DataItems)

        Returns:
            dict: record in self.data. Either main record `self.data[name]` or
            record of a restriction `self.data[name]['restrictions'][restriction_id]`
        """

        assert name in self.data, f"DataItem with name '{name}' is not registered"

        if restriction_id is None:
            if no_additional_restrictions:
                assert len(key_restrictions) == 0, \
                    f"Unexpected restrictions are provided: {repr(key_restrictions)}"
                
            return self.data[name]
        else:
            assert restriction_id in self.data[name]['restrictions'], \
                f"Restrictions with ID={restriction_id} of DataItem '{name}' is not registered"

            item_record = self.data[name]['restrictions'][restriction_id]
            
            # TODO Check carefully!!!
            assert all(item in key_restrictions.items() for item in item_record['fixed_keys'].items()), \
                "Restrictions provided do not conform to base restrictions imposed on restricted DataItem " \
                f"({repr(item_record['fixed_keys'])} != {repr(key_restrictions)})"
            if no_additional_restrictions:
                assert key_restrictions == item_record['fixed_keys'], \
                    "Unexpected restrictions are provided " \
                    f"({repr(item_record['fixed_keys'])} != {repr(key_restrictions)})"
            
            return item_record

    @staticmethod
    def _get_available_index(df: pd.DataFrame) -> int:
        index = 0
        while index in df.index:
            index += 1
        return index

    def index_element(self, keys: dict, index=None, **request) -> None:
        """Add element to indexing table of requested DataItem

        Args:
            index (Any): index of the element of the dataitem
            keys (dict): values of keys for the new element
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted
        """

        item_record = self._access_item_record(**request)
        item = item_record['item']
        
        check_request = {
            key: value
            for key, value in request.items()
            if key != 'key_restrictions'
        }
        raw_restrictions = {**keys, **request['key_restrictions']}
        
        final_restrictions = {
            key: value
            for key, value in raw_restrictions.items()
            if key in item.public_keys or key in item.restrictions or \
                (key in item.additional_keys and item.additional_keys[key] == ColumnType.UNIQUE)
        }
        check_request['key_restrictions'] = final_restrictions

        short_df = self._request_table(**check_request)
        assert len(short_df) <= 1
        drop_item = len(short_df) == 1

        if drop_item: # Drop old item to overwrite with new index
            assert item_record['item'].allow_overwrite
            old_index = short_df.index[0]

            # Assertions for constant keys
            old_keys = short_df.loc[old_index].to_dict()
            old_const_keys = {
                key: value
                for key, value in old_keys.items()
                if key in item.additional_keys and item.additional_keys[key] == ColumnType.CONSTANT
            }

            for key, old_value in old_const_keys.items():
                assert raw_restrictions[key] == old_value

            # Assertion if index is constant
            if index is not None and item.index_expectation == ColumnType.CONSTANT:
                assert index == old_index

            # Remove the row that has to be overwritten
            item_record['table'].drop(old_index, inplace=True)
        
        if index is None:
            assert item.index_expectation != ColumnType.CONSTANT
            index = self._get_available_index(item_record['table'])
        
        item_record['table'].loc[index] = keys
        self.log.debug(f"Item {item_record['name']} indexed element {index} - {repr(keys)}")

        if request['name'] != self._internal('modtimes'):
            modtime_control = item_record['item'].modtime_control
            assert modtime_control is not None

            do_not_record = False
            if modtime_control == 'manual':
                modtime = time.time()
            elif isinstance(modtime_control, dict) and \
                        len(modtime_control) == 1 and \
                        'column' in modtime_control:
                pathname = keys[modtime_control['column']]
                if os.path.isdir(pathname) or os.path.isfile(pathname):
                    modtime = os.path.getmtime(pathname)
                else:
                    do_not_record = True
            else:
                raise Exception(f'Unexpected format of modtime_control: {modtime_control}')
            
            if not do_not_record:
                self.index_element(
                    {
                        self._internal('itemname'): request['name'],
                        self._internal('restriction_id'): request['restriction_id'] 
                            if request['restriction_id'] is not None else np.nan,
                        self._internal('element_index'): index,
                        self._internal('modtime'): modtime,
                    },
                    **self._get_modtimes_request()
                )
            
        
    def _request_table(self, index_restrictions: list=None, **request) -> pd.DataFrame:
        """Get table of elements for given request

        Args:
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted

        Returns:
            pd.DataFrame: the requested dataframe
        """

        item_record = self._access_item_record(**request)
        df = item_record['table']

        if index_restrictions is not None:
            df = df[df.index.isin(index_restrictions)]
        
        if len(request['key_restrictions']) == 0:
            return df
        else:
            restrictions = [
                df[key] == value
                for key, value in request['key_restrictions'].items()
            ]
            total_restriction = restrictions[0]
            for item in restrictions[1:]:
                total_restriction = total_restriction & item
            return df[total_restriction]

    def _assign_table(self, new_df: pd.DataFrame, **request) -> None:
        """Reassign table to some DataItem record.

        Args:
            new_df (pd.DataFrame): the new table (dataframe)
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted
        """

        item_record = self._access_item_record(**request, no_additional_restrictions=True)
        
        # Not doing any extensive checks of `new_df` since this method is not expected
        # to be called from outside of DataStorage
        item_record['table'] = new_df

    def unbind_restriction(self,
            name: str,
            key_restrictions: dict,
            restriction_id: int = None
        ) -> None:
        """Unregister a restriction

        Args:
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted

        Raises:
            RuntimeError: If unbinding of a non-restricted DataItem is requested
        """
        
        if restriction_id is None:
            raise RuntimeError('Cannot unbind DataItem that is not a restriction of another DataItem')
        
        # This is just to go over a few assertions impelemented in _access_item_record
        self._access_item_record(
            name=name,
            key_restrictions=key_restrictions,
            restriction_id=restriction_id,
            no_additional_restrictions=True
        )

        del self.data[name]['restrictions'][restriction_id]
        self.remove_elements(elements_data=[
                {'keys': {
                    self._internal('itemname'): name,
                    self._internal('restriction_id'): restriction_id,
                }}
            ],
            allow_multiple_rows=True,
            **self._get_modtimes_request()
        )

    def process_request(self,
            return_value: str = 'keys',
            expect_single_element: bool = False,
            **request
        ):
        """_summary_

        Args:
            return_value (str, optional): _description_. Defaults to 'keys'.
            expect_single_element (bool, optional): _description_. Defaults to False.

        Returns:
            dict of key-value pairs:
                when return_value=='keys' & expect_single_element
            
            value of the index:
                when return_value=='index' & expect_single_element
            
            (value of the index, dict of key-value pairs):
                when return_value=='keys&index' & expect_single_element

            list of dicts of key-value pairs:
                when return_value=='keys' & not expect_single_element
            
            list of index values:
                when return_value=='index' & not expect_single_element
                
            {index_value: {key-value pairs}}:
                when return_value=='keys&index' & not expect_single_element
        """
        

        df = self._request_table(**request)

        assert return_value in ('keys', 'index', 'keys&index')
        if expect_single_element:
            assert len(df) > 0, f"Element corresponding to request '{repr(request)}' not found"
            assert len(df) == 1, f"Multiple elements corresponding to request '{repr(request)}' were found"
        
            if return_value == 'keys':
                df_dict = df.to_dict(orient='index')
                index, keys = next(iter(df_dict.items()))
                return keys
            elif return_value == 'index':
                return df.index[0]
            else: # 'keys&index'
                df_dict = df.to_dict()
                return df_dict
        else:
            if return_value == 'keys':
                df_dict = df.to_dict(orient='records')
                return df_dict
            elif return_value == 'index':
                return df.index.to_list()
            else: # 'keys&index'
                df_dict = df.to_dict()
                return df_dict

    def rename_indices(self, index_mapping: dict, **request):
        """Rename indices of the table specified by **request.

        Args:
            index_mapping (dict): index mapping for renaming: keys=old_index, value=new_index
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted
        """

        df = self._request_table(**request)
        df = df.rename(index=index_mapping)
        self._assign_table(df, **request)

    def access_key_table(self, **request) -> pd.DataFrame:
        """Access a table that matches indices with values of all keys within current DataItem

        Returns:
            pd.DataFrame: dataframe (NOT a copy!) that stores indices, keys and they values
        """

        return self._access_item_record(
            **request,
            no_additional_restrictions=True
        )['table']

    def unique_keyvalues(self, keyname: str, /, **request) -> list:
        """Get unique values of some key within given DataItem

        Args:
            keyname (str): name of the key
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted

        Returns:
            list: list of unique values of the key
        """

        df = self._request_table(**request)
        return df[keyname].unique().tolist()

    def iter(self, **request):
        """Iterate over rows of a table correspoing to the requested DataItem

        Returns:
            Iterable: returning {'index': index, 'keys': dict of key-value pairs}
        """
        return ItemTableIterator(self, **request)

    def __iter__(self):
        return StoredItemsIterator(self)

    def check_item(self, **request) -> bool:
        """Check if given request specifies any existing element.
        Only complete sets of key values must be specified

        Args:
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted

        Returns:
            bool: if any element was found
        
        Raises:
            AssertionError: when incomplete set of keys was provided
        """
        df = self._request_table(**request)
        assert len(df) == 1 or len(df) == 0, \
            f"Unexpected length of df = {len(df)}"
        return len(df) == 1
    
    def contains_index(self, index, **request) -> bool:
        """Check if a row with given index is present in the table of requested DataItem

        Args:
            index (Any): index of interest
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted

        Returns:
            bool: if row with a given index was found
        """

        df = self._request_table(**request)
        return index in df.index
    
    def _get_dirs_request(self) -> dict:
        return {
            'name': self._internal('directories'),
            'key_restrictions': {},
            'restriction_id': None,
        }
    
    def _get_modtimes_request(self) -> dict:
        return {
            'name': self._internal('modtimes'),
            'key_restrictions': {},
            'restriction_id': None,
        }

    def _get_containing_diritem(self, path: str=None, itemname: str=None):
        assert path is not None or itemname is not None
        assert path is None or itemname is None

        dirs_request = self._get_dirs_request()
        if path is not None:
            dirs_request['key_restrictions'] = {
                self._internal('path'): path,
            }
        else:
            dirs_request['key_restrictions'] = {
                self._internal('itemname'): itemname,
            }

        found_items = self.process_request(return_value='keys', **dirs_request)

        if len(found_items) == 0:
            return None
        
        assert len(set((keys[self._internal('dirobject')] for keys in found_items))) == 1, \
            f"Multiple DirectoryItems describe the same path/itemname"
        dir_object = found_items[0][self._internal('dirobject')]
        return dir_object

    def _request_base_directory(self, dirname: str, create_diritem: bool=False, **request) -> None:
        if create_diritem:
            # Check if appropriate object is already registered
            dir_object = self._get_containing_diritem(path=dirname)

            if dir_object is None:
                # If no DirItems describing the requested dirname were found,
                # the create a new one
                diritem_name = self._internal('dirtemplate').format(itemname=request['name'])
                self.new_item(
                    name=diritem_name,
                    item_type='dirs',
                    mask=dirname,
                    **self.additional_kwargs
                )
                dir_object = diritem_name
        else:
            dir_object = self._internal('directories')

        self.index_element(
            {
                self._internal('path'): dirname,
                self._internal('itemname'): request['name'],
                self._internal('dirobject'): dir_object,
            },
            **self._get_dirs_request()
        )
    
    def number_of_elements(self, **request) -> int:
        """Get the number of elements within required DataItem

        Args:
            **request: `name` - the name of DataItem,
            `restriction_id`, optional - ID of the DataItem restriction if needed
            `key_restrictions`, optional - values of restricted keys if item is restricted

        Returns:
            int: number of elements
        """
        df = self._request_table(**request)
        return df.shape[0]

    def __contains__(self, item_name: str) -> bool:
        assert item_name not in DataStorage.INTERNALS['items'], \
            f"The item name '{item_name}' is protected"
        return item_name in self.data

    def __getattr__(self, item_name: str):
        """Allows to access items as `storage.item_name`

        Args:
            item_name (str): name of the DataItem

        Returns:
            DataItem or child: The object of DataItem or child class

        Raises:
            AssertionError: Name of nonregistered item or some unknown attrubute is provided
        """

        assert item_name in self.data, f"Item with name '{item_name}' is not registered"
        return self.data[item_name]['item']

    def propose_storage_directory(self, item_name: str) -> str:
        abspath = os.path.join(self.wd, f"{item_name}_data")
        relpath = self.preprocess_wd(abspath)
        return relpath

    def preprocess_wd(self, provided_wd: str) -> str:
        if provided_wd == '.':
            return '.'
        
        provided_wd = os.path.normpath(provided_wd)
        if os.path.isabs(provided_wd): # Absolute path
            assert Path(provided_wd).is_relative_to(Path(self.wd)), \
                f"PathItems must be stored inside '{self.wd}' which is the storage for project files"
        elif len(os.path.split(provided_wd)) > 1: # Relative path
            provided_wd = os.path.abspath(os.path.join(self.wd, provided_wd))
            # provided_wd = os.path.abspath(provided_wd)
            
        relpath = os.path.relpath(provided_wd, self.wd)
        if relpath != '.':
            relpath = os.path.join('.', relpath)
        return relpath

    def preprocess_mask(self, mask: str) -> tuple:
        if len(os.path.split(mask)[0]) > 0:
            mask_path = Path(mask)
            return mask_path.name, self.preprocess_wd(str(mask_path.parent))
        else:
            return mask, None

    def postprocess_path(self, provided_relpath: str, absolute: bool=False) -> str:
        if not absolute:
            assert not os.path.isabs(provided_relpath)
            return os.path.normpath(os.path.join(self.wd, provided_relpath))
        else:
            assert Path(provided_relpath).is_relative_to(Path(self.wd)), \
                f"Provided {provided_relpath} is not relative to {self.wd}"
            return provided_relpath

    def verify_parent_directory(self, name: str, **kwargs):
        diritem = self.get_containing_diritem(name=name)
        self.verify_directory(diritem=diritem, **kwargs)

    def verify_directory(self, diritem, key_restrictions: dict, **kwargs):
        diritem_name = diritem.name

        if diritem_name == self._internal('directories'):
            dir_keyvalues = {}
        else:
            dir_keyvalues = {
                key: value
                for key, value in key_restrictions.items()
                if key in diritem.public_keys
            }
        dirname = diritem.get_path(**dir_keyvalues)

        if not os.path.isdir(dirname):
            self.log.info(f"Creating directory '{dirname}'")
            os.mkdir(dirname)

            if diritem_name != self._internal('directories'):
                diritem.include_element(
                    filename=dirname,
                    **dir_keyvalues
                )
        else:
            self.log.debug(f"Directory '{dirname}' already exists")
    
    def get_containing_diritem(self, name: str, key_restrictions: dict={}, **kwargs):
        diritem_name = self._get_containing_diritem(itemname=name)
        assert diritem_name is not None, f"Cannot find DirItem containing the item '{name}'"
        
        res_item = self.data[diritem_name]['item']
        diritem_restriction = {
            key: value
            for key, value in key_restrictions.items()
            if key in res_item.public_keys
        }
        if len(diritem_restriction) == 0:
            return res_item
        else:
            return res_item.get_restricted(keys=diritem_restriction)

    @staticmethod
    def entry_extends(entry, test_entry):
        return all(
            key in test_entry and test_entry[key] == value
            for key, value in entry.items()
        )

    # This is the crucial logic of Transformators and has to be implemented elsewhere
    def key_combinations(self, items: list, nonmerged_keys: list) -> list:
        split_entries = {}
        for itemname in items:
            item = self.data[itemname]['item']
            kv_pairs = item.get_keyvalue_entries()
            split_entries[itemname] = [{
                    key: value
                    for key, value in kv_pair.items()
                    if key in nonmerged_keys
                }
                for kv_pair in kv_pairs
            ]

        # TODO Additional checks??
        
        united_entries = []
        for entries_sublist in split_entries.values():
            for entry in entries_sublist:
                if entry not in united_entries:
                    united_entries.append(entry)
        delete_entries = []
        for i, entry in enumerate(united_entries):
            for j, test_entry in enumerate(united_entries):
                if i != j and DataStorage.entry_extends(entry, test_entry):
                    delete_entries.append(i)
                    break
        united_entries = [
            entry
            for i, entry in enumerate(united_entries)
            if i not in delete_entries
        ]

        check_lengths = set(len(x) for x in united_entries)
        assert len(check_lengths) == 1, f"Inconsitencies in nonmerged key combinations:\n{repr(united_entries)}"
        return united_entries

    def clear_table(self, **request) -> None:
        df = self._request_table(**request)
        
        # Remove all rows from the table to clean it
        df.drop(df.index, inplace=True)

    def deallocate_directories(self, 
            name: str,
            key_restrictions: dict,
            restriction_id: int=None,
            **kwargs
        ) -> None:
        
        assert restriction_id is None, f"Cannot deallocate directories of restriced item '{name}'"
        
        diritem = self.get_containing_diritem(name=name)
        diritem_name = diritem.name
        
        if diritem_name != self._internal('directories'):
            diritem.cleanup()
    
    def remove_elements(self, elements_data, allow_multiple_rows: bool=False, **request) -> None:
        df = self._request_table(**request)

        remove_indices = []
        for element in elements_data:
            if 'index' in element and 'keys' in element:
                index = element['index']
                keys = element['keys']
                df_keys = df.loc[index].to_dict()
                assert df_keys == keys
                remove_indices.append(index)
            elif 'keys' in element:
                keys = element['keys']
                if len(keys) > 0:
                    restrictions = [
                        df[key] == value
                        for key, value in keys.items()
                    ]
                    total_restriction = restrictions[0]
                    for item in restrictions[1:]:
                        total_restriction = total_restriction & item
                    sub_df = df[total_restriction]
                else:
                    sub_df = df
                if not allow_multiple_rows:
                    assert len(sub_df) == 1
                    index = sub_df.index[0]
                    remove_indices.append(index)
                else:
                    remove_indices += sub_df.index.to_list()
            elif 'index' in element:
                remove_indices.append(index)
            else:
                raise RuntimeError
        df.drop(remove_indices, inplace=True)

    # TODO Implement file lookups

    @staticmethod
    def _internal_item(itemname: str) -> str:
        """To avoid conflicts between actual item names and
        internal item names using within DataStorage, we
        augment them with '_internalitem_' prefix.

        Args:
            itemname (str): meaningful name of internal item

        Returns:
            str: augmented name of internal item
        """
        return f'_internalitem_{itemname}'
    
    @staticmethod
    def _internal_key(keyname: str) -> str:
        """To avoid conflicts between actual key names and
        internal key names using within DataStorage, we
        augment them with '_internalkey_' prefix.

        Args:
            keyname (str): meaningful name of internal key

        Returns:
            str: augmented name of internal key
        """
        return f'_internalkey_{keyname}'

    def _internal(self, name: str) -> str:
        return self.INTERNALS_TOTAL[name]

    @staticmethod
    def _fstring_to_wildcard(fstring: str) -> str:
        subs = {
            t[1]: '*'
            for t in string.Formatter().parse(fstring)
            if t[1] is not None
        }
        return fstring.format(**subs)


DataStorage.INTERNALS = {
    'items': {
        'dirtemplate': DataStorage._internal_item('dirtemplate_{itemname}'),
        **{
            itemname: DataStorage._internal_item(itemname)
            for itemname in [
                # DirectoryItem: Index of all directories used to store FileItems and children, etc.
                'directories',
                'modtimes',
            ]
        }
    },
    'keys': {
        'path': PathItem.PATH_KEY,
        **{
            keyname: DataStorage._internal_key(keyname)
            for keyname in [
                # self-explanatory (for instance, 'directories' item stores wd's
                # of all items, thus, the need for 'itemname' key).
                # 'dirobject' is the DirectoryItem that handles self.wd of 'itemname'.
                'itemname',
                'dirobject',

                # For 'modtimes': 'itemname' again, and
                # 'element_index' - element of the item for which mod time is recorded
                # and 'modtime' itself
                'restriction_id',
                'element_index',
                'modtime',
            ]
        },
    },
}
DataStorage.INTERNALS_TOTAL = {
    key: value
    for internal_dict in DataStorage.INTERNALS.values()
    for key, value in internal_dict.items()
}


class ItemTableIterator:
    def __init__(self, base_instance, /, **request):
        df: pd.DataFrame = base_instance._request_table(**request)
        self.main_iter = df.iterrows()

    def __next__(self) -> dict:
        index, keys = next(self.main_iter)
        return {
            'index': index,
            'keys': keys.to_dict(),
        }


class StoredItemsIterator:
    def __init__(self, base_instance):
        self.main_iter = iter(base_instance.data.items())
        self.protected_items = [
            base_instance._fstring_to_wildcard(itename_mask)
            for itename_mask in base_instance.INTERNALS['items'].values()
        ]

    @staticmethod
    def contains_wildcard(value, mask_container):
        return any(
            fnmatch.fnmatch(value, current_mask)
            for current_mask in mask_container
        )

    def __next__(self) -> dict:
        item_name, keys = next(self.main_iter)
        while self.contains_wildcard(item_name, self.protected_items):
            item_name, keys = next(self.main_iter)
        return (item_name, keys['item'])



import time

from ..storage import DataStorage
from chemscripts.mylogging import createLogger


def run_worker():
    logger = createLogger("Worker")

    data_storage = DataStorage({
            'job_directories': {
                'type': 'dirs', 'mask': './workflows/{workflow_name}/{job_id}_work'
            },

            'job_submissions': {
                'type': 'file', 'mask': './workflows/{workflow_name}/{job_id}_submission.json'
            },

            'requests': {
                'type': 'file', 'mask': './requests/{request_id}_request.json'
            },

            'request_responses': {
                'type': 'file', 'mask': './requests/{request_id}_response.json'
            },

            'status_files': {
                'type': 'file', 'mask': './status/{workflow_name}_{job_id}.json'
            },
        },
        logger=logger,
        allow_overwrite=True,
        instantiate_prototypes=True,
        wd='./worker'
    )

    # path = data_storage.requests.get_path(request_id='lol')
    # path = data_storage.request_responses.get_path(request_id='lol')
    # ic(path)

    # Remove older files
    data_storage.job_directories.cleanup()
    data_storage.job_submissions.cleanup()
    data_storage.requests.cleanup()
    data_storage.request_responses.cleanup()
    data_storage.status_files.cleanup()

    time.sleep(1)
    # This will ensure the existence of './workflows', './requests', './status' directories
    for dirobject in (
            data_storage.job_directories.containing_dir().containing_dir(),
            data_storage.requests.containing_dir(),
            data_storage.status_files.containing_dir(),
        ):
        dirobject.ensure_directory()
    

    for name, item in data_storage.data.items():
        ic(name)
        ic(item['table'].sort_index().reset_index(drop=True))


def run_local_worker():
    run_worker() # TODO Run in a separate thread



import string
import glob
import os
import re

from .decorators import register_dataitem
from .abstract_path import PathItem

from chemscripts.utils import str_to_type


@register_dataitem('file')
class FileItem(PathItem):
    def __init__(self,
            files: list = None,
            key_requests: dict = {},
            skip_autoindexing: bool = False,
            **kwargs
        ):
        """

        Args:
            mask (str): filename mask of the form like '{keyA}_{keyB}.ext'.
            wd (str): directory where all FileItem files are stored.
            files (List[str], optional): restrictions to load only specific files
            key_requests (dict, optional): key restrictions for files loading.
            Cannot be provided together with `files` kwarg.
        """
        super().__init__(**kwargs)

        # Will be used to finish initialization in the `checkin` method
        self.checkin_kwargs = {
            'files': files,
            'key_requests': key_requests,
            'skip_autoindexing': skip_autoindexing,
        }

    def checkin(self) -> None:
        """Since this part of initialization requires interaction with `self.storage`
        it is executed a bit after the __init__ call
        """
        super().checkin()

        if not self.checkin_kwargs['skip_autoindexing']:
            self._index_elements(
                key_requests=self.checkin_kwargs['key_requests'],
                required_elements=self.checkin_kwargs['files'],
            )

    def cleanup(self) -> None:
        if self.restriction_id is None:
            # If this is not a restricted version,
            # files will be deleted on cleanup call
            for filename, _ in self:
                self.log.info(f"Removing '{filename}'")
                os.remove(filename)

        # This will remove the table elements corresponding to non-existing paths
        super().cleanup()

    def validate_path(self, path: str) -> bool:
        return os.path.isfile(path)



import copy
import pandas as pd

from .decorators import register_dataitem    
from ..utils import get_logger_shortcuts
from .keynames_layers import GroundKeynamingLayer, KeynamingLayer

from enum import Enum

class ColumnType(Enum):
    # Same value set of UNIQUE keys implies overwriting
    # Example: all keys that are defined by outside user,
    # or itemnames within internal 'dirs' and 'modtimes' tables
    UNIQUE = 1

    # This one is constant for given set of UNIQUE keys
    # Implies the corresponding assertions when overwriting
    # Example: Paths of PathItems, 'dirobject' in 'dirs' table
    CONSTANT = 2

    # Is not required to be constant for given set of UNIQUE keys
    # Example: 'modtime' or all currently used index columns
    VARIABLE = 3

@register_dataitem('None')
class DataItem:
    """
    This is an abstract class, parent for all DataItems.
    However, this is not always an abstract class - it can be used as 'None' dataitem
    It's basic properties are:
    (1) It is always known/discovered/computed
    (2) It has no keys

    TODO WHAT ABOUT DataItems with no keys and one element?
    """

    def __init__(self,
            name: str,
            storage,
            index_expectation: ColumnType,
            restrictions: dict = {},
            restriction_id: int = None,
            additional_keys: dict = {},
            logger=None,
            allow_overwrite: bool = False,
            keys_control = None,
            **kwargs):
        
        assert len(kwargs) == 0, f"DataItem {name} recieved some unprocessed " \
            f"kwargs={repr(tuple(key for key in kwargs.keys()))}"

        self.name = name
        self.storage = storage
        self.restrictions = restrictions
        self.restriction_id = restriction_id

        self.logger = logger
        self.log = get_logger_shortcuts(logger)
        self.allow_overwrite = allow_overwrite
        self.keys_control = keys_control

        self.constructor_kwargs = {}
        if logger is not None:
            self.constructor_kwargs['logger'] = logger
        self.constructor_kwargs['allow_overwrite'] = allow_overwrite
        
        assert isinstance(additional_keys, dict)
        self.additional_keys = additional_keys
        self.index_expectation = index_expectation
        assert self.index_expectation is not ColumnType.UNIQUE # Either CONSTANT or VARIABLE are supported
        
        # Main storage request is used to access the main table of the DataItem even if `self` is restricted
        self.main_storage_request = {
            'name': self.name,
        }
        # Basic storage request is used to access the table of the restriced DataItem if appropriate
        self.storage_request = {
            'name': self.name,
            'key_restrictions': self.restrictions,
        }
        # if self.restriction_id is not None:
        self.storage_request['restriction_id'] = self.restriction_id

        self.log.debug(f"Initialized {self.name}")

    def _restriction_request(self, key_restrictions: dict) -> dict:
        """Get a request to storage object that includes additional restrictions

        Args:
            key_restrictions (dict): additional restrictions

        Returns:
            dict: updated request to storage object
        """

        for key, value in key_restrictions.items():
            if key in self.restrictions:
                assert value == self.restrictions[key], \
                    f"Provided keys conflict with restricted keys of currect DataItem: " \
                    f"{repr(key_restrictions)} vs. {repr(self.restrictions)}"
                
        request = copy.copy(self.storage_request)
        request['key_restrictions'] = {
            **key_restrictions,
            **request['key_restrictions']
        }
        return request

    def include_element(self, keys: dict, index = None) -> None:
        """Store new index with its corresponding values of keys.
        This should be called only from child methods

        Args:
            index (Any): index of the new element
            keys (dict): values of all keys
        """

        assert set(keys.keys()) == set(self._key_names), \
            "Passed key names disagree with existing keys of this DataItem. " \
            f"Passed={list(keys.keys())}, Expected={self._key_names}"
        
        for key, value in keys.items():
            if key in self.restrictions:
                assert value == self.restrictions[key], \
                    f"Provided keys conflict with restricted keys of currect DataItem: " \
                    f"{repr(keys)} vs. {repr(self.restrictions)}"
        keys = {
            **keys,
            **{
                key: value
                for key, value in self.restrictions.items()
            }
        }
        self.storage.index_element(index=index, keys=keys, **self.storage_request)

    def _get_index(self, keys: dict):
        """Get index within storage table by values of keys

        Args:
            keys (dict): values of all keys

        Returns:
            Any: the retrieved object

        Raises:
            AssertionError: on incomplete set of provided keys or multiple objects matched
        """

        assert self.non_empty, \
            f"Container {self.name} is empty, can not get object. Restrictions = {repr(self.restrictions)}"

        assert set(keys.keys()) == set(self.public_keys), \
            f'Not all keys were specified. Passed={list(keys.keys())}, Expected={self.public_keys}'
        
        return self.storage.process_request(
            return_value='index',
            expect_single_element=True,
            **self._restriction_request(keys)
        )

    def _get_indices(self) -> list:
        """Assemble a list of indices of all elements

        Returns:
            list: the resulting list
        """
        return self.storage.process_request(
            return_value='index',
            **self.storage_request
        )

    def _get_keys(self, index=None, **keys) -> dict:
        """Get keys of some item element by its index

        Args:
            index (Any): index of the element

        Returns:
            dict: dict of all keys and values
        """
        
        return self.storage.process_request(
            return_value='keys',
            index_restrictions=([index] if index is not None else None),
            expect_single_element=True,
            **self._restriction_request(key_restrictions=keys)
        )

    def _conforms_keys_control(self, keys: dict) -> bool:
        return self.keys_control is None or self.keys_control(keys)

    def get_restricted(self, keys: dict, constructor_kwargs: dict={}):
        """Build restriction of current DataItem. This does not copy
        any data besides the table of assignments index <-> key values

        Args:
            keys (dict): values of keys used for restriction
            constructor_kwargs (dict): keywords provided by child method `get_restricted`
            that are specific for the exact DataItem being created
            (i.e. cannot be provided via `self.constructor_kwargs`)

        Returns:
            DataItem: the new restricted version of the `self` DataItem that
            has already been registered in the storage and assigned self.restriction_id
        """

        assert self._key_mapping.is_ground

        return self.storage.new_restricted_item(
            **self._restriction_request(keys),
            constructor_kwargs={
                # Passes stuff like self.logger
                **self.constructor_kwargs,

                # Passes what child method `self.get_restricted` has specifically requested
                # like self.keys, self.filename_mask, etc.
                **constructor_kwargs,
            }
        )

    def _rename_indices(self, index_mapping: dict) -> None:
        """Rename indices using `index_mapping`

        Args:
            index_mapping (dict): provided index mapping: keys=old_indices, values=new_indices
        """
        self.storage.rename_indices(index_mapping, **self.storage_request)

    def __iadd__(self, other) -> None:
        """This just implements a few assertions that are
        universal for all DataItem childred. The core expectation
        is that `other` must be a restriction of `self`. The process of
        concatenation must be implemented within children classes
        using pattern like this:
        ```
        super().__iadd__(other)
        for index, keys, some_data in other:
            self.include_element(index=index, keys=keys, some_data=some_data)
        ```

        Args:
            other (DataItem child): other DataItem - a restriction of `self`
            containing some new elements that must be included into `self`.
            Most likely, the `other` object will soon be removed via `other.retire()`
        """

        assert self.name == other.name, \
            f"Merging FileItems with different names: FileItem({repr(self.name)}) += FileItem({repr(other.name)})"
        
        assert self._key_mapping.is_ground # No keys are currently renamed

        if not self.allow_overwrite:
            for index, keys in other._base_iter():
                assert not self.contains_keys(keys), \
                    f"Main DataItem already contains element with {repr(keys)}"

    def contains_keys(self, keys: dict) -> bool:
        """Check if given set of keys specifies any element
        in the DataItem's table.

        Args:
            keys (dict): values of ALL keys of this DataItem

        Returns:
            bool: True (exactly one element found),
            False (no elements found), AssertionError otherwise
        """
        
        assert set(keys.keys()) == set(self._key_names), \
            "Passed key names disagree with existing keys of this DataItem. " \
            f"Passed={list(keys.keys())}, Expected={self._key_names}"
        return self.storage.check_item(**self._restriction_request(keys))

    def contains_index(self, index) -> bool:
        """Check if a row with given index is present in this DataItem

        Args:
            index (Any): index of interest

        Returns:
            bool: True (found element with `index`), False (didn't find an element)
        """
        return self.storage.contains_index(index, **self.storage_request)

    def retire(self) -> None:
        """All restrictions of DataItems are meant to have limited lifetime, so
        at some point they are unbound from the storage in order to be garbage collected
        """
        
        assert self.restriction_id is not None, \
            f"Attempting to remove DataItem that is not a restriction"
        self.storage.unbind_restriction(**self.storage_request)

    def access_key_table(self) -> pd.DataFrame:
        """Access a table that matches indices with values of all keys within current DataItem

        Returns:
            pd.DataFrame: dataframe (NOT a copy!) that stores indices, keys and they values
        """
        
        return self.storage.access_key_table(**self.storage_request)

    def __len__(self) -> int:
        return len(self.access_key_table())

    def get_values_of_key(self, keyname: str) -> list:
        """Summarize all values that a given key has within this DataItem

        Args:
            keyname (str): the key to summarize

        Returns:
            List[str]: list of all existing values of this key
        
        Raises:
            AssertionError: when incorrect key name is provided
        """

        assert keyname in self._key_names, f'Incorrect key {keyname} was provided'
        return self.storage.unique_keyvalues(keyname, **self.storage_request)

    def get_keyvalue_entries(self) -> list:
        pub_keys = self.public_keys
        return [
            {
                key: value
                for key, value in keys.items()
                if key in pub_keys
            }
            for index, keys in self._base_iter()
        ]

    def cleanup(self) -> None:
        self.storage.clear_table(**self.storage_request)

    def _base_iter(self, **settings):
        """This iterator factory is meant to be used within
        all DataItem child classes 

        Returns:
            DataItemIterator: iterator that exposes indices, key-value pairs
            of elements and custom data supplied from child classes
        """
        return DataItemIterator(self, request=self.storage_request, **settings)

    def __iter__(self):
        raise RuntimeError("Cannot iterate over DataItems"
                           "(use a child class like ObjectItems or FileItems)")

    def checkin(self) -> None:
        """Placeholder of a method that will be overloaded in DataItem children
        that need to interact with self.storage to finish initialization (e.g. FileItem)
        """
        pass

    def with_renamed_keys(self, all_keys_mapping: dict):
        new_keys_mapping = {
            keyname: 
                all_keys_mapping[keyname] if keyname in all_keys_mapping
                else keyname
            for keyname in self.public_keys
        }
        self._key_mapping = KeynamingLayer(
            key_mapping=new_keys_mapping,
            base=self._key_mapping
        )
        return self

    def release_keyname_layer(self) -> None:
        self._key_mapping = self._key_mapping.get_base()

    def __len__(self) -> int:
        return self.storage.number_of_elements(**self.storage_request)

    @property
    def public_keys(self) -> list:
        return self._key_mapping.get_public_keys()

    @property
    def _key_names(self) -> list:
        return list(self.additional_keys.keys())

    @property
    def is_empty(self):
        return len(self) == 0
    
    @property
    def non_empty(self):
        return len(self) > 0


class DataItemIterator:
    def __init__(self,
            base_instance,
            request: dict,
            call=None
        ):
        """Initializes iterator that is inteded to be used for
        all child classes of DataItem 

        Args:
            base_instance (DataItem or child): instance to iterate over
            request (dict): request to self.storage
            calls (None/Callable/Dict, optional): allows to customize
            the third element of the tuple for various children of DataItem.
            Can be a function that returns whatever must be put at the third place, or
            a dict of functions - then there is a dict at the third place
            (key=key of self.calls, value=output of self.calls[key]).
            These custom callables must take {'index': index, 'keys': keys} as the only arg.
        """
        
        self.storage_iter = base_instance.storage.iter(**request)

        assert call is None or callable(call)
        self.call = call

    def __iter__(self):
        return self

    def __next__(self) -> tuple:
        """On each iteration it returns a tuple of two or three element

        Returns:
            tuple: (index: Any, keys: Dict, additional_output: depends on self.calls)
        """
        next_dict = next(self.storage_iter)

        iteration_raw = (next_dict['index'], next_dict['keys'])
        if self.call is None:
            return iteration_raw
        else:
            return self.call(*iteration_raw)



import copy

class KeynamingLayer:
    def __init__(self, key_mapping: dict, base=None):
        self.own_mapping = key_mapping
        self.base = base
        
        # just 'mapping' internal -> external
        # 'inverse mapping' external -> internal
        if self.base is None:
            self.total_mapping = key_mapping
        else:
            self.total_mapping = {
                internal_name: self.own_mapping[base_external_name]
                for internal_name, base_external_name in self.base.total_mapping.items()
            }
        
        self.inverse_total_mapping = {
            value: key
            for key, value in self.total_mapping.items()
        }

        self.is_ground = False
    
    def keyname_to_internal(self, keyname: str) -> str:
        return self.total_mapping[keyname]
    
    def keyname_to_external(self, keyname: str) -> str:
        return self.inverse_total_mapping[keyname]
    
    def map_to_external(self, keys: dict) -> dict:
        return {
            self.total_mapping[key] 
                if key in self.total_mapping 
                    else key 
                : value
            for key, value in keys.items()
        }
    
    def map_to_internal(self, keys: dict) -> dict:
        return {
            self.inverse_total_mapping[key] 
                if key in self.inverse_total_mapping 
                    else key 
                : value
            for key, value in keys.items()
        }
    
    def get_base(self):
        assert self.base is not None
        return self.base

    def get_public_keys(self) -> list:
        return list(key for key in self.total_mapping.values())


class GroundKeynamingLayer(KeynamingLayer):
    def __init__(self, keys: list):
        key_mapping = {
            keyname: keyname
            for keyname in keys
        }
        super().__init__(key_mapping=key_mapping)

        self.is_ground = True
    
    def get_base(self):
        raise RuntimeError('Attempted release of the ground keynaming layer')



import string
import glob
import os
import re
from pathlib import Path

from .decorators import register_dataitem
from .abstract_data import DataItem, ColumnType
from .keynames_layers import GroundKeynamingLayer, KeynamingLayer

from chemscripts.utils import str_to_type


@register_dataitem('path')
class PathItem(DataItem):
    PATH_KEY = '_pathitem_Path'
    def __init__(self,
            mask: str,
            wd: str = None,
            **kwargs
        ):
        """

        Args:
            mask (str): path mask of the form like '{keyA}_{keyB}.ext'.
            wd (str): directory where all FileItem files are stored.
            files (List[str], optional): restrictions to load only specific files
            key_requests (dict, optional): key restrictions for files loading.
            Cannot be provided together with `files` kwarg.
        """
        if 'additional_keys' not in kwargs:
            kwargs['additional_keys'] = {}
        kwargs['additional_keys'][self.PATH_KEY] = ColumnType.CONSTANT
        super().__init__(index_expectation=ColumnType.VARIABLE, **kwargs)

        self.mask, self.wd = self._obtain_paths(
            mask=mask,
            wd=wd,
        )

        self.modtime_control = {'column': self.PATH_KEY}

        self._key_mapping = GroundKeynamingLayer(keys=[key for key in self._keys_iter()])

    def _obtain_paths(self, mask=None, wd=None) -> tuple:
        assert mask is not None
        fixed_mask, mask_wd = self.storage.preprocess_mask(mask)

        if wd is not None:
            provided_wd = self.storage.preprocess_wd(wd)
        
        if wd is not None and mask_wd is not None:
            assert provided_wd == mask_wd, \
                "Mismatch of working directories extracted from mask and wd-kwarg: " \
                f"'{provided_wd}' vs. '{mask_wd}'"

        if mask_wd is not None:
            fixed_wd = mask_wd
        elif wd is not None: # Then, provided_wd is the preprocessed path
            fixed_wd = provided_wd
        else:
            fixed_wd = self.storage.propose_storage_directory(self.name)
        
        if fixed_mask != '.':
            fixed_mask = self.storage.preprocess_wd(os.path.join(fixed_wd, fixed_mask))
        return fixed_mask, fixed_wd

    def checkin(self) -> None:
        """Since this part of initialization requires interaction with `self.storage`
        it is executed a bit after the __init__ call
        """
        if self.wd != '.' or self.mask != '.': # if it is not '_internalitem_directories'
            create_diritem = self.wd != '.'

            self.storage._request_base_directory(
                self.wd,
                create_diritem=create_diritem,
                **self.storage_request
            )

    def _index_elements(self,
            required_elements: list = None,
            key_requests: dict = {}
        ):
        """Load existing files whose names are consistent with
        filename pattern (f-string `self.mask`).
        Record key values for each file.
        
        Args:
            files (List[str], optional): `files` kwarg can be passed to constrain
            parsing to specific files.
            key_requests (dict, optional): kwargs can be passed to constrain parsing
            to some specific key values.

        Raises:
            RuntimeError: on ambiguous/inappropriate filename patterns
        """
        # TODO Record modification time
        
        assert not (required_elements is not None and len(key_requests) > 0), \
            f"Cannot initialize FileItem '{self.name}' with both keys AND filenames specified: " \
            f"keys={repr(key_requests)}, paths={repr(required_elements)}"
            
        # Either include my filename mask or include all specified files
        if required_elements is None:
            # Use filename mask
            search_mask = self._get_wildcard_expression(**key_requests)
            include_elements = list(glob.glob(search_mask))
        else:
            include_elements = required_elements

            # Assert that all requested files exist
            all_files_exist = True
            for file in include_elements:
                if os.path.isfile(file):
                    all_files_exist = False
                    break
            if not all_files_exist:
                raise RuntimeError('Some of the requested files are missing:' + \
                    repr([file for file in include_elements if not os.path.isfile(file)]))

        for file in include_elements:
            try:
                keys = self._get_keyvalues(file)
            except IncompatiblePathError:
                continue

            super().include_element(
                keys={
                    self.PATH_KEY: file,
                    **keys
                }
            )
    
    def validate_path(self, path: str) -> bool:
        return os.path.isfile(path) or os.path.isdir(path)

    def _get_keyvalues(self, filename: str, assert_underscores=True) -> dict:
        """Obtain values of keys from provided filename
        by parsing using self.mask

        Args:
            filename (str): filename to be processed

        Raises:
            IncompatiblePathError: if some parsed key value contains '_' symbol
            usually this indicates that some file of different dataitem '{a}_{b}.ext'
            was parsed as '{c}.ext' which is incorrect

        Returns:
            dict: key-value pairs of a given filename
        """

        if len(list(self._keys_iter())) == 0:
            return {}

        filename = self.storage.preprocess_wd(filename)
        assert self.validate_path(self._postprocess_path(filename)), \
            f"Processed path '{filename}' is invalid for the item '{self.name}'. Make sure that path exists and check your masks"
        
        self.log.debug(f"Parsing key values from '{filename}'")

        # Regex for removing self.wd from name
        # wd_mask = os.path.join(self.wd, '*')
        # wd_regex = re.escape(wd_mask).replace(r"\*", "(.*)")

        # Regex for extracting keys
        pattern = self._get_wildcard_expression(absolute=False)
        regex = re.escape(pattern).replace(r"\*", "(.*)")

        # wd_matches = re.findall(wd_regex, filename)
        # assert len(wd_matches) != 0 and not(len(wd_matches) > 1), \
        #     f"Unable to extract working directory from '{filename}'"
        # pure_name = wd_matches[0]
        pure_name = filename

        assert '*' not in pure_name, "Found * symbol in the name of '{pure_name}'. Don't use it"
        matches = re.findall(regex, pure_name) # TODO Doesn't generate ALL possible matches :((
        assert len(matches) != 0, f"{pure_name} doesn't match the pattern '{pattern}'"
        assert not(len(matches) > 1), f"Several matches found for {pure_name} (has to be only 1). Pattern = '{pattern}'"
        match = matches[0]
        if isinstance(match, str):
            match = (match,)
        
        # Need to check that '_' isn't present in matched parts
        if assert_underscores:
            skip_file = False
            for value in match:
                if '_' in value:
                    self.log.info(f"File '{pure_name}' doesn't match the pattern for "
                            f"{self.name} (key contains '_')")
                    skip_file = True
            if skip_file:
                raise IncompatiblePathError(filename, pattern)
        
        key_value_pairs = {
            key: str_to_type(value)
            for key, value in zip(self._keys_iter(), match)
        }
        
        return key_value_pairs

    def access_element(self, **keys) -> str:
        """Access filename by its full set of keys

        Returns:
            Any: the retrieved filename

        Raises:
            AssertionError: on incomplete set of provided keys or multiple filenames matched
        """
        
        for key in keys.keys():
            assert key in self.public_keys, f"Key '{key}' is not present"
        keys = self._key_mapping.map_to_internal(keys)

        absolute_path = self._get_keys(**keys)[self.PATH_KEY]
        return self._postprocess_path(absolute_path, absolute=True)
        return relative_path

    def include_element(self, filename: str, assert_underscores=True, **keys_provided) -> None:
        """Index a new filename as an element of current FileItem

        Args:
            filename (str): name of the file to be included
        """

        for key in keys_provided.keys():
            assert key in self.public_keys, f"{key} not in {self.public_keys}"
        keys_provided = self._key_mapping.map_to_internal(keys_provided)

        if not self._conforms_keys_control(keys_provided):
            return

        filename = self.storage.preprocess_wd(filename)
        keys_determined = self._get_keyvalues(filename, assert_underscores=assert_underscores)

        if len(keys_provided) > 0:
            assert set(keys_provided.keys()) == set(keys_determined.keys()), \
                "Problem with provided keys: not all (or unexpected) keys were specified. " \
                f"Passed={list(keys_provided.keys())}, Expected={self._key_names}"
            
            assert keys_provided == keys_determined, \
                f"Problem with provided keys: mismatch between filename and provided keys " \
                f"Passed={repr(keys_provided)}, Expected={repr(keys_determined)}"
        
        super().include_element(
            keys={self.PATH_KEY: self._postprocess_path(filename), **keys_determined}
        )

    def get_restricted(self, keys: dict, **kwargs):
        """Get the restricted (filtered) version of the fileitem.
        It's very important to pass only the necessary files into transforms.
        Restricted version can be updated from inside transforms and
        then manually merged into the main fileitem.

        Args:
            keys (Dict[str, Any]): key values used for restriction

        Returns:
            FileItem: restricted version (detached from the `self` fileitem)
        """

        return super().get_restricted(keys,
            constructor_kwargs={
                'mask': self._get_mask(**keys),
                **kwargs
            }
        )

    def __iadd__(self, other):
        """Merge with another FileItem

        Args:
            other (FileItem): Usually, restricted version of FileItem
            produced by transform is merged into the main FileItem

        Returns:
            FileItem: updated main FileItem
        """
        
        # This only performs a few assertions
        super().__iadd__(other)

        for filename, keys in other:
            self.include_element(filename=filename, **keys, **other.restrictions)
        return self

    def _process_iteration(self, index, keys):
        return self._postprocess_path(keys[PathItem.PATH_KEY]), \
            {
                key: value
                for key, value in keys.items()
                if key != PathItem.PATH_KEY
            }

    def __iter__(self):
        pub_keys = self.public_keys
        return self._base_iter(
            call=lambda index, keys: (
                keys[PathItem.PATH_KEY],
                {
                    key: value
                    for key, value in self._key_mapping.map_to_external({
                        key: value
                        for key, value in keys.items()
                        if key != PathItem.PATH_KEY
                    }).items()
                    if key in pub_keys
                }
            )
        )
    
    def private_iterator(self):
        return self._base_iter()
        # return self._base_iter(
        #     call=lambda index, keys: self._process_iteration(index, keys)
        # )

    def get_path(self, **keys):
        assert set(self.public_keys) == set(keys.keys()), \
            "Mismatch between provided and expected keys: " \
            f"Provided={repr(list(keys.keys()))}, Expected={repr(self.public_keys)}"
        
        keys = self._key_mapping.map_to_internal(keys)

        if self.mask != '.':
            self.storage.verify_parent_directory(**self._restriction_request(key_restrictions=keys))

        return self._postprocess_path(self.mask.format(**keys))

    def _get_mask(self, **kwargs):
        """Get the mask of filename with some keys set to specified values.
        Pass kwargs to set some of the keys to specific values.

        Returns:
            str: the resulting filename mask
        """

        # General substitution of all keys to '*'
        subs = {
            key: '{%s}' % key
            for key in self._keys_iter()
        }
        
        # Replace some '*' with specified values
        for key, value in kwargs.items():
            subs[key] = value
        
        return self.mask.format(**subs)

    def _get_wildcard_expression(self, absolute=True, **kwargs):
        """Get the wildcard expression of filename by replacing all keys with '*' symbol.
        Pass kwargs to set some of the keys to specific values.

        Returns:
            str: the resulting filename mask
        """

        # General substitution of all keys to '*'
        subs = {
            key: '*'
            for key in self._keys_iter()
        }
        
        # Replace some '*' with specified values
        for key, value in kwargs.items():
            subs[key] = value

        if absolute:        
            return self._postprocess_path(self.mask.format(**subs))
        else:
            return self.mask.format(**subs)

    
    def _postprocess_path(self, filename: str, absolute: bool=False) -> str:
        # TODO This doesn't work when self.wd has '{stuff}'
        # assert Path(filename).is_relative_to(Path(self.wd))
        return self.storage.postprocess_path(filename, absolute=absolute)

    def containing_dir(self):
        return self.storage.get_containing_diritem(**self.storage_request)

    def cleanup(self) -> None:
        if self.restriction_id is None:
            self.storage.deallocate_directories(**self.storage_request)

        remove_elements = []
        for index, keys in self._base_iter():
            path = keys[PathItem.PATH_KEY]
            if not (os.path.isdir(path) or os.path.isfile(path)):
                remove_elements.append({'index': index, 'keys': keys})
        self.storage.remove_elements(remove_elements, **self.storage_request)

    @staticmethod
    def _keys_as_list(mask) -> list:
        return [
            t[1]
            for t in string.Formatter().parse(mask)
            if t[1] is not None
        ]

    def _keys_iter(self) -> str:
        """Lazy iterator over keys of the filename mask

        Yields:
            str: formatting elements of self.mask
        """

        for t in string.Formatter().parse(self.mask):
            if t[1] is not None:
                yield t[1]
    
    @property
    def _key_names(self) -> list:
        """Create a list of all keys of the current FileItem

        Returns:
            List[str]: resulting list
        """
        return [ key for key in self._keys_iter() ] + super()._key_names


class IncompatiblePathError(Exception):
    def __init__(self, path, mask):
        self.message = f"Path '{path}' does not match the mask '{mask}'"
        super().__init__(self.message)



from enum import Enum

class TransformResult(Enum):
    FINISHED = 0
    FAILED = 1
    REPEAT = 2
    LATER = 3


class TransformState:
    def __init__(self,
            result: TransformResult,
            comeback_method: str=None,
            internals: dict={}
        ) -> None:

        self.result = result

        if self.result == TransformResult.FAILED or self.result == TransformResult.FINISHED:
            assert len(internals) == 0 and comeback_method is None, \
                f"Cannot repeat transform execution when the last result is '{repr(self.result)}'"
        
        self.internals = internals
        self.comeback_method = comeback_method
    
    @property
    def is_final(self):
        return self.result == TransformResult.FAILED or self.result == TransformResult.FINISHED

    @staticmethod
    def returnvalue_to_state(return_value):
        if isinstance(return_value, TransformState):
            return return_value
        else:
            assert return_value in TransformStateFactories.RETURN_VALUES, \
                f"'{return_value}' is not a valid return value from a transform"
            return TransformStateFactories.RETURN_VALUES[return_value]()


class TransformStateFactories:
    def transform_finished():
        return TransformState(TransformResult.FINISHED)

    def transform_failed():
        return TransformState(TransformResult.FAILED)

    def transform_repeat(comeback: str=None, **internals):
        return TransformState(TransformResult.REPEAT, comeback_method=comeback, internals=internals)
        
    def transform_later(comeback: str=None, **internals):
        return TransformState(TransformResult.LATER, comeback_method=comeback, internals=internals)

    RETURN_VALUES = {
        None: transform_finished,
        'done': transform_finished,
        'failed': transform_failed,
    }



class InstructionInfo(dict):
    """This thing allows access via attibutes (info.type = ...)

    Args:
        dict (_type_): _description_
    """
    def __init__(self, input={}):
        super(InstructionInfo, self).__init__(input)

    def __getattr__(self, name):
        return self.__getitem__(name)

    def __setattr__(self,name,value):
        self.__setitem__(name,value)


class Instruction:
    def __init__(self, data: dict):
        self.info = InstructionInfo(data)
        self.finished = False

    def activate_dependents(self, stack) -> None:
        if 'activates' not in self.info:
            return

        for instruction_name in self.info.activates:
            activated = False
            for instruction in stack.instructions:
                if instruction_name == instruction.info.name:
                    assert not activated
                    assert self.info.name in instruction.info.activated_by
                    index = instruction.info.activated_by.index(self.info.name)
                    del instruction.info.activated_by[index]
                    stack.log.info(f"Stack {repr(self.info.name)}: Activated instruction '{instruction.info.name}'")
                    activated = True
            assert activated
            
    @property
    def is_subframe(self):
        return self.info.type == 'subframe'

    @property
    def is_active(self):
        return not self.finished and (not 'activated_by' in self.info or len(self.info.activated_by) == 0)



import os, sys
import shutil
import time
import json
import copy
import ntpath
import types
import string
import inspect
import tarfile
from typing import List
import numpy as np
import pandas as pd

from autopy import TransformStateFactories as ret
from autopy import TransformState
from .abstract_transform import Transform

from chemscripts.excelutils import ExcelSheet
from chemscripts.nbo.isosurfaces import generate_isosurface
from chemscripts.nbo import NBO3LogParser, NBO6LogParser, NboSymmMatrix, generate_reorder_map
from chemscripts import utils, nbo
from chemscripts import fchkutils
IRC_SKIP_POINTS = 40

NBO_PAIRING = {
    'dmmo': 419,
    'dmnbo': 420,
    'fmo': 699,
    'fnbo': 69,
    'aomo': 228,
    'aonbo': 288,
}


item_to_str = lambda i: i if isinstance(i, str) else i.name

# def if_check(condition, result, other=None):
#     if condition:
#         return (result,)
#     elif other is None:
#         return tuple()
#     else:
#         return (other,)


def nonblocking_subprocess(name,
        input,
        output,
        command_prepare,
        output_process,
        aware_keys=[],
        merged_keys=[],
        nproc=1,
        calcdir=None,
    ):
    if not isinstance(input, list):
        input = [input]
    if not isinstance(output, list):
        output = [output]

    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(x) for x in input],
        TARGET_ITEMS = [item_to_str(x) for x in output],
        AWARE_KEYS = aware_keys,
        MERGED_KEYS = merged_keys,
    )

    def RunSubprocess_exec(thread_manager, **kw):
        full_kwargs = {**kw}
        if calcdir is not None:
            id, current_calcdir = thread_manager.prepare_new_job()
            full_kwargs[calcdir] = current_calcdir
        else:
            id = thread_manager.prepare_new_job(use_calcdir=False)
        full_kwargs['id'] = id

        execution_command = command_prepare(**full_kwargs)

        thread_manager.run_nonblocking(
            id=id,
            command=execution_command,
            nproc=nproc
        )

        return ret.transform_later(comeback='wait', id=id)
    
    def RunSubprocess_wait(thread_manager, id, **kw):
        if not thread_manager.is_finished(id):
            return ret.transform_later(comeback='wait', id=id)
        else:
            return ret.transform_repeat(comeback='finish', id=id)

    def RunSubprocess_finish(thread_manager, id, **kw):
        full_kwargs = {'id': id, **kw}
        if calcdir is not None:
            current_calcdir = thread_manager.id_to_calcdir(id)
            full_kwargs[calcdir] = current_calcdir
        output_process(**full_kwargs)
        thread_manager.finalize_task(id)
        return ret.transform_finished()
    
    res.set_method({
        'exec': RunSubprocess_exec,
        'wait': RunSubprocess_wait,
        'finish': RunSubprocess_finish,
    })
    
    return res


def function_to_script(f, subs={}):
    defaults = {
        'module_dir': os.getcwd()
    }
    script_text = inspect.getsource(f)
    script_lines = script_text.split('\n')[1:]

    starting_spaces = None
    for i, line in enumerate(script_lines):
        if len(line.replace(' ', '')) == 0:
            continue
        cur_spaces = len(line) - len(line.lstrip())
        if starting_spaces is None or starting_spaces > cur_spaces:
            starting_spaces = cur_spaces

    for i, line in enumerate(script_lines):
        script_lines[i] = script_lines[i][starting_spaces:]
    
    replacements = set()
    for i, line in enumerate(script_lines):
        if 'INSERT_HERE' not in line:
            continue

        assert line.count('=') == 1
        var_name = line.split('=')[0].strip()
        assert var_name in subs or var_name in defaults, f"Variable '{var_name}' of script template '{f.__name__}' was not defined"
        assert line.split('=')[1].strip() == 'INSERT_HERE'
        
        if var_name in subs:
            value = subs[var_name]
        else:
            value = defaults[var_name]

        script_lines[i] = line.replace('INSERT_HERE', repr(value))
        replacements.add(var_name)
    # assert len(replacements) == len(subs)

    return '\n'.join(script_lines)

def write_py_function(file, f, args, subs={}) -> str:
    assert os.path.isabs(file), f"Path '{file}' must be absolute"

    script_text = function_to_script(f, subs=subs)
    with open(file, 'w') as f:
        f.write(script_text)

    args_text = ' '.join([f'"{x}"' if isinstance(x, str) else f'"{repr(x)}"' for x in args])
    execution_command = f"{sys.executable} {file} {args_text}"
    return execution_command

def get_arg_extension(mapping):
    if isinstance(mapping, dict):
        key_change = mapping
    elif isinstance(mapping, str):
        mapping = '\n'.join([
            line
            for line in mapping.splitlines()
            if not line.lstrip().startswith('#') # Ignore comment lines
        ])
        key_change = {}
        for part in mapping.split():
            sides = part.split('->')
            key_change[sides[1]] = sides[0] # Maps 'parent_key' -> 'child_key'
    
    def extension(input):
        return {
            key if key not in key_change
            else key_change[key]
                : value
            for key, value in input.items()
        }
    return extension

def pyfunction_subprocess(name,
        input,
        output,
        output_process,
        custom_command=None,
        pyfunction=None,
        argv_prepare=None,
        pyfile=None,
        aware_keys=[],
        merged_keys=[],
        calcdir=None,
        subs=None,
        nproc=1,
    ):
    assert pyfile is not None or calcdir is not None
    assert custom_command is not None or \
        (pyfunction is not None and argv_prepare is not None)
    if custom_command is not None:
        assert pyfunction is None and argv_prepare is None
    if subs is not None:
        assert pyfunction is not None
    
    if not isinstance(input, list):
        input = [input]
    if not isinstance(output, list):
        output = [output]
    if pyfile is not None:
        output.append('pyfile')

    arg_renames = []
    if pyfile is not None:
        arg_renames.append(f"{item_to_str(pyfile)}->pyfile")
    arg_extend = get_arg_extension(' '.join(arg_renames))

    def command_prepare(**kw):
        if pyfile is not None:
            pyfile_fname = kw['pyfile'].get_path()
        
        if calcdir is not None:
            script_path = os.path.join(kw[calcdir], 'script.py')
        elif pyfile is not None and calcdir is None:
            script_path = pyfile_fname

        if custom_command is not None:
            execution_command = custom_command(file=script_path, **arg_extend(kw))
        else:
            if subs is not None:
                subs_optional = {'subs': subs(**arg_extend(kw))}
            else:
                subs_optional = {}
            execution_command = write_py_function(
                file=script_path,
                f=pyfunction,
                args=argv_prepare(**arg_extend(kw)),
                **subs_optional
            )

        if pyfile is not None and calcdir is not None:
            shutil.copy2(script_path, pyfile_fname)
        if pyfile is not None:
            kw['pyfile'].include_element(pyfile_fname)

        return execution_command

    def output_process_raw(**kw):
        output_process(**arg_extend(kw))

    template = nonblocking_subprocess(
        'pyfunction_subprocess_template',
        input=input,
        output=output,
        command_prepare=command_prepare,
        output_process=output_process_raw,
        aware_keys=aware_keys,
        merged_keys=merged_keys,
        nproc=nproc,
        calcdir=calcdir,
    )
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = template.specs['SOURCE_ITEMS'],
        TARGET_ITEMS = [item_to_str(x) if x != 'pyfile' else pyfile for x in template.specs['TARGET_ITEMS']],
        AWARE_KEYS = template.specs['AWARE_KEYS'],
        MERGED_KEYS = template.specs['MERGED_KEYS'],
    )
    res.extend(template, ' '.join(arg_renames))
    return res

#
# EXCEL
#
def parse_xlsx(name, input, output, block_name='Main'):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )

    def ParseXlsx_exec(xlsx, df):
        xlsx_path = xlsx.access_element()
        sheet = ExcelSheet()
        sheet.read_xlsx(xlsx_path)
        block = sheet.block(block_name)
        df.include_element(block['data'])
        return ret.transform_finished()

    ParseXlsx_template = Transform(
        NAME = 'ParseXlsxtemplate',
        SOURCE_ITEMS = ['xlsx'],
        TARGET_ITEMS = ['df'],
    )
    ParseXlsx_template.set_method({'exec': ParseXlsx_exec})

    res.extend(ParseXlsx_template, f"""\
        {item_to_str(input)}->xlsx
        {item_to_str(output)}->df
    """)
    return res

def construct_excel_sheet(name, input, block_name, sheet=None, xlsx=None, block_sort_key=None, merged_keys=[]):
    additional_targets = []
    additional_targets_renamed = []
    additional_renames = []
    assert sheet is not None or xlsx is not None

    if sheet is not None:
        sheet_name = item_to_str(sheet)
        additional_targets.append('sheet')
        additional_targets_renamed.append(sheet_name)
        additional_renames.append(f'{sheet_name}->sheet')

    if xlsx is not None:
        xlsx_name = item_to_str(xlsx)
        additional_targets.append('xlsx')
        additional_targets_renamed.append(xlsx_name)
        additional_renames.append(f'{xlsx_name}->xlsx')

    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['data'],
        TARGET_ITEMS = [*additional_targets],
        MERGED_KEYS = merged_keys
    )

    def ConstructExcel_exec(data, **kw):
        excelsheet = ExcelSheet()

        blocks = []
        for block_data, keys in data:
            cur_blockname = block_name(block_data)
            blocks.append({
                'data': block_data,
                'keys': keys,
                'name': cur_blockname,
            })

        if block_sort_key is not None:
            blocks.sort(key=block_sort_key)
        else:
            blocks.sort(key=lambda x: x['name'])

        for full_dict in blocks:
            block_data, keys, cur_blockname = full_dict['data'], full_dict['keys'], full_dict['name']
            column_names = list(set(
                key
                for element in block_data
                for key in element.keys()
            ))
            excelsheet.add_block(
                blockname=cur_blockname,
                cols=column_names
            )
            for element in block_data:
                excelsheet.add_row(blockname=cur_blockname, data=element)
        if sheet is not None:
            kw['sheet'].include_element(excelsheet)
        if xlsx is not None:
            res_fname = kw['xlsx'].get_path()
            excelsheet.save_xlsx(res_fname)
            kw['xlsx'].include_element(res_fname)
        return ret.transform_finished()
    
    template.set_method({'exec': ConstructExcel_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [*additional_targets_renamed],
        MERGED_KEYS = merged_keys
    )
    res.extend(template, f"""\
        {item_to_str(input)}->data
        {' '.join(additional_renames)}
    """)
    return res

#
# DF OPERATIONS
#

# Extract rows
def ExtractDfRows_exec(df, rows):
    df_object = df.access_element()
    assert isinstance(df_object, list)
    row_keys = rows.public_keys
    for item in df_object:
        key_values = {
            key: value
            for key, value in item.items()
            if key in row_keys
        }
        short_item = {
            key: value
            for key, value in item.items()
            if key not in row_keys
        }
        rows.include_element(short_item, **key_values)
    return ret.transform_finished()

ExtractDfRows_template = Transform(
    NAME = 'ExtractDfRowstemplate',
    SOURCE_ITEMS = ['df'],
    TARGET_ITEMS = ['rows'],
)
ExtractDfRows_template.set_method({'exec': ExtractDfRows_exec})

def extract_df_rows(name, input, output):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(ExtractDfRows_template, f"""\
        {item_to_str(input)}->df
        {item_to_str(output)}->rows
    """)
    return res


# Extract specific column
def extract_df_column(name, input, output, column_name):
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['df'],
        TARGET_ITEMS = ['column'],
    )
    def ExtractDfColumn_exec(df, column):
        df_object = df.access_element()
        assert isinstance(df_object, list)
        column_keys = column.public_keys
        for item in df_object:
            key_values = {
                key: value
                for key, value in item.items()
                if key in column_keys
            }
            column.include_element(item[column_name], **key_values)
        return ret.transform_finished()
    template.set_method({'exec': ExtractDfColumn_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(template, f"""\
        {item_to_str(input)}->df
        {item_to_str(output)}->column
    """)
    return res

#
# BASIC DATA OPERATIONS
#

def select(name, input, output, select_method, merged_keys=None):
    if merged_keys is None:
        add_kw = {}
    else:
        add_kw = {'MERGED_KEYS': merged_keys}

    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['data'],
        TARGET_ITEMS = ['selected'],
        **add_kw
    )
    def Select_exec(data, selected):
        selected.include_element(
            select_method([x for x in data])
        )
        return ret.transform_finished()
    template.set_method({'exec': Select_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
        **add_kw
    )
    res.extend(template, f"""\
        {item_to_str(input)}->data
        {item_to_str(output)}->selected
    """)
    return res

def exec(name, input, output, method, note=[], aware_keys=[], merged_keys=[]):
    if isinstance(input, list):
        input_array = [item_to_str(x) for x in input]
    else:
        input_array = [item_to_str(input)]
        
    if isinstance(output, list):
        output_array = [item_to_str(x) for x in output]
    else:
        output_array = [item_to_str(output)]

    if isinstance(note, list):
        note_array = [item_to_str(x) for x in note]
    else:
        note_array = [item_to_str(note)]

    res = Transform(
        NAME = name,
        SOURCE_ITEMS = input_array,
        TARGET_ITEMS = output_array,
        NOTE_ITEMS = note_array,
        AWARE_KEYS = aware_keys,
        MERGED_KEYS = merged_keys
    )

    def method_wrapper(**kw):
        res = method(**kw)
        if res is None:
            return ret.transform_finished()
        else:
            assert isinstance(res, TransformState), f"Value returned by exec method '{method}' is not a TransformState"
            return res

    res.set_method({'exec': method_wrapper})
    return res

def verify_type_consistency(*items):
    assert all(item.is_file for item in items) or all(item.is_object for item in items), \
        f"Inconsistent typing of items: {', '.join(f'item.name' for item in items)}"

def map(name,
        input,
        output,
        mapping, 
        aware_keys=[],
        note_items=[],
        ensure_one_to_one=False, 
        ensure_new_keys=None,
        include_none=True
    ):
    if isinstance(input, list):
        input_array = [item_to_str(x) for x in input]
    else:
        input_array = [item_to_str(input)]
    assert not isinstance(output, list)
    output_item = item_to_str(output)

    res = Transform(
        NAME = name,
        SOURCE_ITEMS = input_array,
        TARGET_ITEMS = [output_item],
        NOTE_ITEMS = note_items,
        AWARE_KEYS = aware_keys
    )
    
    if ensure_new_keys is not None:
        ensure_new_keys = set(ensure_new_keys)

    def Map_exec(**kw):
        add_args = {}
        if kw[output_item].is_file:
            add_args[output_item] = kw[output_item].get_path()
        
        for note_item in note_items:
            assert kw[note_item].is_path, f"Expected path (file/dir) as note item. Got '{note_item}'"
            assert len(kw[note_item].public_keys) == 0, f"Note item '{note_item}' has some unresolved keys: {kw[note_item].public_keys}"
        
        result = mapping(
            **{
                inp_item: kw[inp_item].access_element()
                for inp_item in input_array
            },
            **{
                note_item: kw[note_item].get_path()
                for note_item in note_items
            },
            **{
                aware_key: kw[aware_key]
                for aware_key in aware_keys
            },
            **add_args
        )

        expansion = isinstance(result, types.GeneratorType)
        if expansion:
            assert not ensure_one_to_one, f"Unexpected expansion during map '{name}'"
            for value, keys in result:
                if ensure_new_keys is not None:
                    assert set(keys.keys()) == ensure_new_keys, \
                        f"Mismatch between created keys '{set(keys.keys())}' and expected ones '{repr(ensure_new_keys)}'"
                
                if value is not None or include_none:
                    kw[item_to_str(output)].include_element(value, **keys)
        else:
            if result is not None or include_none:
                kw[item_to_str(output)].include_element(result)

        return ret.transform_finished()
    
    res.set_method({'exec': Map_exec})
    return res

def forward_file(target, input_file=None, keys={}, input=None, copy_method=None):
    assert (input is None) ^ (input_file is None)

    if input is not None:
        old_path = input.access_element(**keys)
    else:
        old_path = input_file

    new_path = target.get_path(**keys)

    if copy_method is not None:
        copy_method(old_path, new_path)
    else:
        shutil.copy2(old_path, new_path)
    
    target.include_element(new_path)

def forward_object(target, input_object=None, keys={}, input=None, copy_method=None):
    assert (input is None) ^ (input_object is None)

    if input is not None:
        old_object = input.access_element(**keys)
    else:
        old_object = input_object

    if copy_method is None:
        new_object = copy.deepcopy(old_object)
    else:
        new_object = copy_method(old_object)

    target.include_element(new_object, **keys)

def forward_any_type(target, input=None, input_element=None, file_copy=None, object_copy=None, **kwargs):
    if input is not None:
        verify_type_consistency(target, input)
    
    if target.is_file:
        forward_file(target=target, input=input, input_file=input_element, copy_method=file_copy, **kwargs)
    elif target.is_object:
        forward_object(target=target, input=input, input_object=input_element, copy_method=object_copy, **kwargs)
    else:
        raise RuntimeError(f"Unknown type of '{target.name}'")


def restrict(name, input, ref, output, merged_keys=[], copy_method=None):
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['input', 'ref'],
        TARGET_ITEMS = ['output'],
        MERGED_KEYS = merged_keys
    )

    def Restrict_exec(input, ref, output, **kw):
        assert not ((input.is_object) ^ (output.is_object)), f"{input.name}.is_object = {input.is_object}, {output.name}.is_object = {output.is_object}"
        assert not ((input.is_file) ^ (output.is_file)), f"{input.name}.is_file = {input.is_file}, {output.name}.is_file = {output.is_file}"
        if copy_method is not None:
            assert input.is_object, "Object copy method is provided but we are not dealing with objects here"

        input_keys = set(input.public_keys)
        ref_keys = set(ref.public_keys)
        output_keys = set(output.public_keys)
        merged_set = set(merged_keys)
        restriction_set = input_keys.intersection(ref_keys).intersection(output_keys).intersection(merged_set)
        assert len(restriction_set) > 0, f"No keys to perform restriction"

        accepted_kvpairs = []
        for _, keys in ref:
            kvpair = {
                key: value
                for key, value in keys.items()
                if key in restriction_set
            }
            if kvpair not in accepted_kvpairs:
                accepted_kvpairs.append(kvpair)
        
        for accepted_kvpair in accepted_kvpairs:
            forward_any_type(
                input=input,
                keys=accepted_kvpair,
                target=output,
                object_copy=copy_method
            )
        return ret.transform_finished()

    template.set_method({'exec': Restrict_exec})
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input), item_to_str(ref)],
        TARGET_ITEMS = [item_to_str(output)],
        MERGED_KEYS = merged_keys
    )
    res.extend(template, f"""\
        {item_to_str(input)}->input
        {item_to_str(ref)}->ref
        {item_to_str(output)}->output
    """)
    return res


def substitute(name, input, substituent, output, merged_keys, copy_method=None):
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['input', 'substituent'],
        TARGET_ITEMS = ['output'],
        MERGED_KEYS = merged_keys
    )

    def Substitute_exec(input, substituent, output, **kw):
        # Primary checks
        verify_type_consistency(input, substituent, output)
        if copy_method is not None:
            assert input.is_object, "Object copy method is provided but we are not dealing with objects here"

        input_keys = set(input.public_keys)
        substituent_keys = set(substituent.public_keys)
        output_keys = set(output.public_keys)
        assert input_keys == substituent_keys and substituent_keys == output_keys, \
            f"Keys mismatch: input_keys='{input_keys}', substituent_keys='{substituent_keys}', output_keys='{output_keys}'"
        
        # Forward all primary elements
        for start_element, keys in input:
            forward_any_type(
                input_element=start_element,
                keys=keys,
                target=output,
                object_copy=copy_method
            )

        # Borrow missing elements from substituent
        input_kvset = [keys for _, keys in input]
        for start_element, keys in substituent:
            if keys not in input_kvset:
                forward_any_type(
                    input_element=start_element,
                    keys=keys,
                    target=output,
                    object_copy=copy_method
                )
        return ret.transform_finished()

    template.set_method({'exec': Substitute_exec})
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input), item_to_str(substituent)],
        TARGET_ITEMS = [item_to_str(output)],
        MERGED_KEYS = merged_keys
    )
    res.extend(template, f"""\
        {item_to_str(input)}->input
        {item_to_str(substituent)}->substituent
        {item_to_str(output)}->output
    """)
    return res

class CustomTransformProxy:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
    
    def _run_factory(self):
        assert 'factory' in self.kwargs, f"Custom transform is missing its factory: '{self.kwargs}'"
        factory = self.kwargs['factory']
        assert callable(factory), f"The transform factory '{factory}' is not callable"
        return factory(**{
            key: value
            for key, value in self.kwargs.items()
            if key != 'factory'
        })

    def _update_kwargs(self, modifying_kwargs: dict):
        for key, value in modifying_kwargs.items():
            self.kwargs[key] = value

    def customize(self, **modifying_kwargs) -> Transform:
        self._update_kwargs(modifying_kwargs)
        return self._run_factory()

    def use_default(self) -> Transform:
        return self._run_factory()

def custom(name, **kw):
    return CustomTransformProxy(name=name, **kw)


#
# RDKIT
#

# Create Mol instance
def RdkitMolFromSmiles_exec(smiles, mol):
    smiles_str = smiles.access_element()
    
    from rdkit import Chem
    params = Chem.SmilesParserParams()
    params.removeHs = False
    params.sanitize = True
    mol_object = Chem.MolFromSmiles(smiles_str, params)
    mol_object = Chem.AddHs(mol_object)

    mol.include_element(mol_object)
    return ret.transform_finished()

RdkitMolFromSmiles_template = Transform(
    NAME = 'RdkitMolFromSmilestemplate',
    SOURCE_ITEMS = ['smiles'],
    TARGET_ITEMS = ['mol'],
)
RdkitMolFromSmiles_template.set_method({'exec': RdkitMolFromSmiles_exec})

def rdkit_from_smiles(name, input, output):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(RdkitMolFromSmiles_template, f"""\
        {item_to_str(input)}->smiles
        {item_to_str(output)}->mol
    """)
    return res

# Perform conformational sampling
def rdkit_sampling(name, input, output, num_tries, postprocessing=None):
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['mol'],
        TARGET_ITEMS = ['conformers'],
    )

    def RdkitSampling_exec(mol, conformers):
        from rdkit import Chem
        from rdkit.Chem import AllChem
        from ringo import Confpool

        mol_object = mol.access_element()
        assert isinstance(mol_object, Chem.rdchem.Mol)
        p = Confpool()

        for conformer_idx in range(num_tries):
            AllChem.EmbedMolecule(mol_object)

            geom = np.zeros((mol_object.GetNumAtoms(), 3))
            for i in range(mol_object.GetNumAtoms()):
                pos = mol_object.GetConformer().GetAtomPosition(i)
                geom[i, 0] = pos.x
                geom[i, 1] = pos.y
                geom[i, 2] = pos.z
            p.include_from_xyz(geom, f"Conformer {conformer_idx}")

        p.atom_symbols = [atom.GetSymbol() for atom in mol_object.GetAtoms()]
        if postprocessing is not None:
            postprocessing(p)

        conformers.include_element(p)
        return ret.transform_finished()
    template.set_method({'exec': RdkitSampling_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(template, f"""\
        {item_to_str(input)}->mol
        {item_to_str(output)}->conformers
    """)
    return res

#
# CREST
#

# Perform conformational sampling
def crest_sampling(name, input, output, level, nproc=None):
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['start'],
        TARGET_ITEMS = ['conformers_xyz'],
    )

    def CrestSampling_exec(start, thread_manager, **kw):
        start_data = start.access_element()
        xyz, syms = start_data['xyz'], start_data['sym']

        id, calcdir = thread_manager.prepare_new_job()
        utils.write_xyz(xyz, syms, os.path.join(calcdir, 'start.xyz'))
        if nproc is None:
            if thread_manager.maxproc < 8:
                nproc_local = thread_manager.maxproc
            else:
                nproc_local = 8
        else:
            nproc_local = nproc

        thread_manager.run_nonblocking(id, f'exec_crest.sh {calcdir} start.xyz {level} -T {nproc_local}', nproc_local)

        return ret.transform_later(comeback='wait', id=id)
    
    def CrestSampling_wait(thread_manager, id, **kw):
        if not thread_manager.is_finished(id):
            return ret.transform_later(comeback='wait', id=id)
        else:
            return ret.transform_repeat(comeback='finish', id=id)
    
    def CrestSampling_finish(conformers_xyz, thread_manager, id, **kw):
        calcdir = thread_manager.id_to_calcdir(id)
        conformers_file = os.path.join(calcdir, 'crest_conformers.xyz')
        assert os.path.isfile(conformers_file), f"Cannot find file '{conformers_file}'"

        final_xyz_name = conformers_xyz.get_path()
        shutil.copy2(conformers_file, final_xyz_name)
        conformers_xyz.include_element(final_xyz_name)

        thread_manager.finalize_task(id)
        return ret.transform_finished()
    
    template.set_method({
        'exec': CrestSampling_exec,
        'wait': CrestSampling_wait,
        'finish': CrestSampling_finish,
    })
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(template, f"""\
        {item_to_str(input)}->start
        {item_to_str(output)}->conformers_xyz
    """)
    return res


#
# Confpool
#

# Save Confpool
def SaveConfpool_exec(confpool, xyz):
    p = confpool.access_element()
    xyzname = xyz.get_path()
    p.save(xyzname)
    xyz.include_element(xyzname)
    return ret.transform_finished()

SaveConfpool_template = Transform(
    NAME = 'SaveConfpooltemplate',
    SOURCE_ITEMS = ['confpool'],
    TARGET_ITEMS = ['xyz'],
)
SaveConfpool_template.set_method({'exec': SaveConfpool_exec})

def save_confpool(name, input, output):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(SaveConfpool_template, f"""\
        {item_to_str(input)}->confpool
        {item_to_str(output)}->xyz
    """)
    return res

# Load Confpool
def load_confpool(name, input, output, postprocessing=None, merged_keys=None, load_keys=[]):
    if merged_keys is None:
        add_kw = {}
    else:
        add_kw = {'MERGED_KEYS': merged_keys}
    
    LoadConfpool_template = Transform(
        NAME = 'LoadConfpooltemplate',
        SOURCE_ITEMS = ['xyz'],
        TARGET_ITEMS = ['confpool'],
        **add_kw
    )

    def LoadConfpool_exec(xyz, confpool, **kw):
        from ringo import Confpool
        p = Confpool()
        atom_symbols = None
        for elem, _ in xyz:
            if isinstance(elem, str):
                p.include_from_file(elem)
            elif isinstance(elem, dict):
                if atom_symbols is None:
                    atom_symbols = elem['sym']
                else:
                    assert atom_symbols == elem['sym']
                p.include_from_xyz(elem['xyz'], '')
                for key in load_keys:
                    assert key in elem
                    p[len(p) - 1][key] = float(elem[key])
            else:
                raise TypeError(f"Cannot process type '{type(elem)}'")
        if atom_symbols is not None:
            p.atom_symbols = atom_symbols
        
        if postprocessing is not None:
            postprocessing(p)
        
        confpool.include_element(p)
        return ret.transform_finished()
    
    LoadConfpool_template.set_method({'exec': LoadConfpool_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
        **add_kw
    )
    res.extend(LoadConfpool_template, f"""\
        {item_to_str(input)}->xyz
        {item_to_str(output)}->confpool
    """)
    return res

# Merge Confpools
def merge_confpools(name, inputs, output, postprocessing=None):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(x) for x in inputs],
        TARGET_ITEMS = [item_to_str(output)],
    )

    def MergeConfpools_exec(**kw):
        from ringo import Confpool
        p = Confpool()
        atom_symbols = None
        for inp_item in inputs:
            for source_p, _ in kw[inp_item]:
                if atom_symbols is not None:
                    assert atom_symbols == source_p.atom_symbols
                else:
                    atom_symbols = source_p.atom_symbols
                
                for m in source_p:
                    p.include_from_xyz(m.xyz, m.descr)
        p.atom_symbols = atom_symbols

        if postprocessing is not None:
            postprocessing(p)

        kw[output].include_element(p)
        return ret.transform_finished()
    res.set_method({'exec': MergeConfpools_exec})
    return res

# Split Confpool
def split_confpool(name, input, output, keygen):
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['confpool'],
        TARGET_ITEMS = ['structs'],
    )

    def SplitConfpool_exec(confpool, structs):
        p = confpool.access_element()

        keys = structs.public_keys
        assert len(keys) == 1
        key = keys[0]
        
        for m in p:
            structs.include_element({
                    'xyz': m.xyz,
                    'sym': p.atom_symbols,
                },
                **{key: keygen(m)}
            )
        return ret.transform_finished()
    template.set_method({'exec': SplitConfpool_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(template, f"""\
        {item_to_str(input)}->confpool
        {item_to_str(output)}->structs
    """)
    return res

def xyz_to_file(name, input, output, merged_keys=[], description=None, sort_key=None):
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['geom'],
        TARGET_ITEMS = ['xyzfile'],
        MERGED_KEYS = merged_keys
    )

    def XyzToFile_exec(geom, xyzfile):
        from ringo import Confpool
        p = Confpool()
        atom_symbols = None
        geom_elements = [x for x in geom]
        if sort_key is not None:
            geom_elements.sort(key=sort_key)

        for data, keys in geom_elements:
            if description is not None:
                kwargs = {
                    **keys,
                    **{
                        key: value
                        for key, value in data.items()
                        if key not in ('xyz', 'sym')
                    }
                }
                p.include_from_xyz(data['xyz'], description(**kwargs))
            else:
                p.include_from_xyz(data['xyz'], **keys)
            
            if atom_symbols is not None:
                assert atom_symbols == data['sym']
            else:
                atom_symbols = copy.copy(data['sym'])
                p.atom_symbols = data['sym']
        xyzname = xyzfile.get_path()
        p.save(xyzname)
        xyzfile.include_element(xyzname)
        return ret.transform_finished()
    
    template.set_method({'exec': XyzToFile_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
        MERGED_KEYS = merged_keys
    )
    res.extend(template, f"""\
        {item_to_str(input)}->geom
        {item_to_str(output)}->xyzfile
    """)
    return res

#
# GAUSSIAN
#

def xyz_from_gausslog(logname, get_scf=False, get_qh=False, keep_name=True):
    xyz, sym = utils.parse_gaussian_log(logname)
    res = {
        'xyz': xyz,
        'sym': sym,
    }
    if keep_name:
        res['log'] = logname
    if get_scf:
        res['scf'] = utils.get_gaussian_scfener(logname)
    if get_qh:
        res['qhG'] = utils.get_goodvibes_g(logname)
    return res

def load_gauss_scan(logname: str, gen_key, get_scf: bool=False, get_variable: bool=True):
    scan_data = utils.parse_gaussian_scan(logname=logname, keep_energy=get_scf, get_variable=get_variable)
    for i, data_item in enumerate(scan_data):
        res = {
            'xyz': data_item['xyzs'],
            'sym': data_item['syms'],
            'var': data_item['variable'],
        }
        if get_scf:
            res['scf'] = float(data_item['scfener'])
        yield res, gen_key(index=i, data=res)

def _irc_include_point(xyz, sym, energy, rxcoord, get_scf: bool=False, get_rxcoord: bool=False, get_parameters=None):
    res = {
        'xyz': xyz,
        'sym': sym,
    }
    if get_scf:
        res['scf'] = energy
    if get_rxcoord:
        res['rxcoord'] = rxcoord
    if get_parameters is not None:
        from ringo import Confpool
        p = Confpool()
        p.include_from_xyz(res['xyz'], '')
        p.atom_symbols = res['sym']
        res = {
            **res,
            **get_parameters(p[0])
        }
    return res

def load_gauss_irc(logname: str, gen_key, get_start: bool=True, **kwargs):
    irc_data = utils.parse_irc(logname=logname, verify=False)
    irc_points = irc_data['points']
    
    if get_start:
        res = _irc_include_point(xyz=irc_points['start'], sym=irc_data['syms'], energy=irc_points['start_energy'], rxcoord=0.0, **kwargs)
        yield res, gen_key(data=res, index=0)
        
    for index, (xyz, rxcoord, energy) in enumerate(zip(irc_points['f_geoms'], irc_points['f_rxcoord'], irc_points['f_energy'])):
        res = _irc_include_point(xyz=xyz, sym=irc_data['syms'], energy=energy, rxcoord=rxcoord, **kwargs)
        yield res, gen_key(data=res, index=index+1)

    for index, (xyz, rxcoord, energy) in enumerate(zip(irc_points['b_geoms'], irc_points['b_rxcoord'], irc_points['b_energy'])):
        res = _irc_include_point(xyz=xyz, sym=irc_data['syms'], energy=energy, rxcoord=rxcoord, **kwargs)
        yield res, gen_key(data=res, index=-index-1)


def prepare_gjf(gjftext: str, remove_chk, remove_oldchk):
    lines = gjftext.splitlines()
    kwline_index = None
    for i, line in enumerate(lines):
        if line.startswith('#'):
            kwline_index = i
            break
    assert kwline_index is not None
    for i in reversed(range(kwline_index)):
        if lines[i].startswith('%chk=') and remove_chk or \
           lines[i].startswith('%oldchk=') and remove_oldchk:
            del lines[i]
    return '\n'.join(lines)

def fstring_key_iter(input: str) -> List[str]:
    for t in string.Formatter().parse(input):
        if t[1] is not None:
            yield t[1]


def run_gaussian(name,
        inpgeom,
        gjf_template,
        log,
        output=None,
        output_method=None,
        gjf=None,
        chk=None,
        oldchk=None,
        additional_files=None, # raw generated file -> fileitem
        nproc=None,
        **kwargs
    ):
    assert utils.check_availability('exec_gaussian.sh')

    additional_sources = []
    additional_sources_renamed = []
    additional_targets = []
    additional_targets_renamed = []
    additional_renames = []

    if gjf is not None:
        additional_targets.append('gjf')
        additional_targets_renamed.append(item_to_str(gjf))
        additional_renames.append(f'{item_to_str(gjf)}->gjf')
    if chk is not None:
        additional_targets.append('chk')
        additional_targets_renamed.append(item_to_str(chk))
        additional_renames.append(f'{item_to_str(chk)}->chk')
    if oldchk is not None:
        additional_sources.append('oldchk')
        additional_sources_renamed.append(item_to_str(oldchk))
        additional_renames.append(f'{item_to_str(oldchk)}->oldchk')
    if output is not None:
        additional_targets.append('output')
        additional_targets_renamed.append(item_to_str(output))
        additional_renames.append(f'{item_to_str(output)}->output')
    
    if additional_files is not None:
        additional_files_str = {key: item_to_str(value) for key, value in additional_files.items()}
        for file in additional_files_str.values():
            additional_targets.append(file)
            additional_targets_renamed.append(file)
    else:
        additional_files_str = {}

    execute_gaussian = any(x is not None for x in (chk, oldchk, output, additional_files))
    assert execute_gaussian, f"Seems like you didn't request Gaussian execution (just gjf), but this is not implemented yet"

    def init_gaussian(inpstruct, calcdir, id, **kw):
        default_charge = 0
        default_mult = 1
        charge = default_charge
        mult = default_mult

        data = inpstruct.access_element()
        if isinstance(data, dict):
            xyzs, syms = data['xyz'], data['sym']
            coord_text = utils.to_xyz(xyzs, syms, include_header=False)
        elif isinstance(data, str) and data.endswith('.sdf'):
            from chemscripts.geom import Molecule
            mol = Molecule(sdf=data)
            charge = mol.total_charge()
            xyzs, syms = mol.as_xyz()
            coord_text = utils.to_xyz(xyzs, syms, include_header=False)
        else:
            raise ValueError(f"Unable to start Gaussian from '{repr(data)}'")

        template = open(gjf_template, 'r').read()
        template_keys = list(fstring_key_iter(template))
        current_kwargs = {}
        if charge != default_charge or 'chrg' in template_keys:
            current_kwargs['chrg'] = charge
        if mult != default_mult or 'mult' in template_keys:
            current_kwargs['mult'] = mult
        if 'nproc' in template_keys:
            assert nproc is not None, f"GJF template '{gjf_template}' requires 'nproc' to be provided"
            current_kwargs['nproc'] = nproc
        for key in current_kwargs.keys():
            assert key in template_keys, f"Template '{gjf_template}' "\
                f"does not support non-default values of '{key}'"

        gjf_string = template.format(
            xyz=coord_text,
            **current_kwargs,
            **kwargs
        )
        gjf_string = prepare_gjf(gjf_string, remove_chk=chk is None, remove_oldchk=oldchk is None)
        gjfname = os.path.join(calcdir, 'start.gjf')
        with open(gjfname, 'w') as f:
            f.write(gjf_string)
        
        if gjf is not None:
            gjf_backup_name = kw['gjf'].get_path()
            with open(gjf_backup_name, 'w') as f:
                f.write(gjf_string)
            kw['gjf'].include_element(gjf_backup_name)

        if oldchk is not None:
            oldchk_name = kw['oldchk'].access_element()
            shutil.copy2(oldchk_name, os.path.join(calcdir, 'old.chk'))
        
        return f'exec_gaussian.sh {calcdir} start.gjf {id}'

    def finish_gaussian(log, calcdir, **kw):
        conformers_file = os.path.join(calcdir, 'start.log')
        assert os.path.isfile(conformers_file), f"Cannot find file '{conformers_file}'"
        assert utils.is_normal_termination(conformers_file, '.gjf'), f"Failed '{conformers_file}'"

        final_log_name = log.get_path()
        shutil.copy2(conformers_file, final_log_name)
        log.include_element(final_log_name)

        if output is not None:
            kw['output'].include_element(output_method(final_log_name))

        if chk is not None:
            final_chk_name = kw['chk'].get_path()
            chkname = os.path.join(calcdir, 'current.chk')
            assert os.path.isfile(chkname), f"Cannot find file '{chkname}'"
            shutil.copy2(chkname, final_chk_name)
            kw['chk'].include_element(final_chk_name)

        for raw_filename, fileitem in additional_files_str.items():
            final_name = kw[fileitem].get_path()
            raw_fullname = os.path.join(calcdir, raw_filename)
            assert os.path.isfile(raw_fullname), f"Cannot find file '{raw_fullname}'"
            shutil.copy2(raw_fullname, final_name)
            kw[fileitem].include_element(final_name)
    
    if execute_gaussian:
        template = nonblocking_subprocess(
            name='RunGaussian_template',
            input=['inpstruct', *additional_sources],
            output=['log', *additional_targets],
            command_prepare=init_gaussian,
            output_process=finish_gaussian,
            calcdir='calcdir',
            nproc=nproc
        )
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(inpgeom), *additional_sources_renamed],
        TARGET_ITEMS = [item_to_str(log), *additional_targets_renamed],
    )
    res.extend(template, f"""\
        {item_to_str(inpgeom)}->inpstruct
        {item_to_str(log)}->log
        {' '.join(additional_renames)}
    """)
    return res


# Load gaussian logs independently from Gaussian calculation itself
def load_gaussian_logs(name, input, output, parse_method, **kwargs):
    return map(
        name=name,
        input=input,
        output=output,
        mapping=parse_method,
        **kwargs
    )


# chk->fchk conversion
def init_formchk(chk, calcdir, **kw):
    chk_fname = chk.access_element()
    shutil.copy2(chk_fname, os.path.join(calcdir, 'mol.chk'))
    return f'exec_gaussian_formchk.sh {calcdir} mol.chk'

def finish_formchk(fchk, calcdir, **kw):
    fchk_fname = os.path.join(calcdir, 'mol.fchk')
    assert os.path.isfile(fchk_fname), f"Cannot find file '{fchk_fname}'"
    final_fchk_name = fchk.get_path()
    shutil.copy2(fchk_fname, final_fchk_name)
    fchk.include_element(final_fchk_name)

def formchk_gaussian(name, input, output):
    assert utils.check_availability('exec_gaussian_formchk.sh')

    template = nonblocking_subprocess(
        name=name,
        input=['chk'],
        output=['fchk'],
        command_prepare=init_formchk,
        output_process=finish_formchk,
        calcdir='calcdir',
        nproc=1
    )
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(template, f"""\
        {item_to_str(input)}->chk
        {item_to_str(output)}->fchk
    """)
    return res


# fchk->chk conversion
def init_unfchk(fchk, calcdir, **kw):
    fchk_fname = fchk.access_element()
    shutil.copy2(fchk_fname, os.path.join(calcdir, 'mol.fchk'))
    return f'exec_gaussian_unfchk.sh {calcdir} mol.fchk'

def finish_unfchk(chk, calcdir, **kw):
    chk_fname = os.path.join(calcdir, 'mol.chk')
    assert os.path.isfile(chk_fname), f"Cannot find file '{chk_fname}'"
    final_chk_name = chk.get_path()
    shutil.copy2(chk_fname, final_chk_name)
    chk.include_element(final_chk_name)

def unfchk_gaussian(name, input, output):
    assert utils.check_availability('exec_gaussian_unfchk.sh')

    template = nonblocking_subprocess(
        name=name,
        input=['fchk'],
        output=['chk'],
        command_prepare=init_unfchk,
        output_process=finish_unfchk,
        calcdir='calcdir',
        nproc=1
    )
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )
    res.extend(template, f"""\
        {item_to_str(input)}->fchk
        {item_to_str(output)}->chk
    """)
    return res

# NBO6 execution

# emplates.run_nbosix('nbo6_calc',
#                 input='nbogauss_47', output='nbo6_out', additional_output={'dmmo': 'nbo6_dmmo', 'fmo': 'nbo6_fmo'}
#             ),

NBO_MATRIX_CONVENTIONS = {
    'dmmo': 419,
    'dmnbo': 420,
    'fmo': 699,
    'fnbo': 69,
    'aomo': 228,
    'aonbo': 288,
}

def run_nbosix(name,
        input,
        output=None,
        modified_input=None,
        matrix_files=None,
        nbo_keywords=[],
        **kwargs
    ):
    assert utils.check_availability('exec_nbo6.sh')

    additional_sources = []
    additional_sources_renamed = []
    additional_targets = []
    additional_targets_renamed = []
    additional_renames = []

    if output is not None:
        additional_targets.append('output')
        additional_targets_renamed.append(item_to_str(output))
        additional_renames.append(f'{item_to_str(output)}->output')

    if modified_input is not None:
        additional_targets.append('modified_input')
        additional_targets_renamed.append(item_to_str(modified_input))
        additional_renames.append(f'{item_to_str(modified_input)}->modified_input')

    generated_matrices = []
    for matrix_type, matrix_item in matrix_files.items():
        generated_matrices.append(matrix_type)
        additional_targets.append(matrix_type)
        additional_targets_renamed.append(item_to_str(matrix_item))
        additional_renames.append(f'{item_to_str(matrix_item)}->{matrix_type}')

    execute_nbo = (output is not None) or (len(matrix_files) > 0)
    assert execute_nbo, f"Seems like you didn't request NBO execution (just modified input), but this is not implemented yet"

    def init_nbosix(input, calcdir, id, **kw):
        inpfile = input.access_element()
        assert inpfile.endswith('.47'), f"Input for NBO6 must end with '.47'. Got this: '{inpfile}'"

        input_lines = open(inpfile, 'r').readlines()
        assert input_lines[1] == ' $NBO  $END\n', f"Unexpected start of file '{inpfile}'. Needed ' $NBO  $END'"
        keywords = [
            *nbo_keywords,
            *(
                f"{matrix_type.upper()}=W{NBO_MATRIX_CONVENTIONS[matrix_type]}" # For example, 'DMMO=W419'
                for matrix_type in generated_matrices
            )
        ]
        joined_keywords = ' '.join(keywords)
        assert '\n' not in joined_keywords, f"Unexpected linebreak found: '{joined_keywords}'"
        input_lines[1] = f' $NBO {joined_keywords} $END\n'
        input_worker_name = os.path.join(calcdir, 'calc.47')
        with open(input_worker_name, 'w') as f:
            f.write(''.join(input_lines))

        if modified_input is not None:
            modinput_backup_name = kw['modified_input'].get_path()
            with open(modinput_backup_name, 'w') as f:
                f.write(input_lines)
            kw['modified_input'].include_element(modinput_backup_name)
        
        return f'exec_nbo6.sh {calcdir} calc.47'

    def finish_nbosix(calcdir, **kw):
        output_worker_file = os.path.join(calcdir, 'NBO.OUT')
        assert os.path.isfile(output_worker_file), f"Cannot find file '{output_worker_file}'"
        assert utils.nbo_check_normal_termination(output_worker_file), f"Failed '{output_worker_file}'"

        if output is not None:
            final_output_name = kw['output'].get_path()
            shutil.copy2(output_worker_file, final_output_name)
            kw['output'].include_element(final_output_name)

        for matrix_type in generated_matrices:
            matrix_worker_file = os.path.join(calcdir, f'FILE.{NBO_MATRIX_CONVENTIONS[matrix_type]}')
            assert os.path.isfile(matrix_worker_file), f"Cannot find file '{matrix_worker_file}'"
            final_matrix_name = kw[matrix_type].get_path()
            shutil.copy2(matrix_worker_file, final_matrix_name)
            kw[matrix_type].include_element(final_matrix_name)
    
    if execute_nbo:
        template = nonblocking_subprocess(
            name='RunNBO6_template',
            input=['input', *additional_sources],
            output=[*additional_targets],
            command_prepare=init_nbosix,
            output_process=finish_nbosix,
            calcdir='calcdir',
            nproc=1
        )
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input), *additional_sources_renamed],
        TARGET_ITEMS = [*additional_targets_renamed],
    )
    res.extend(template, f"""\
        {item_to_str(input)}->input
        {' '.join(additional_renames)}
    """)
    return res


# Load NBO results (parser and matrices)
def load_nbo_results(name,
        input,
        parser,
        dmnbo=None,
        fnbo=None,
        dmmo=None,
        fmo=None,
        nbo_version=3
    ):
    assert nbo_version == 3 or nbo_version == 6, f"Only NBO3/Gaussian and NBO6 are supported"

    additional_sources = []
    additional_targets = []
    if dmnbo is not None:
        additional_sources.append(item_to_str(dmnbo[0]))
        additional_targets.append(item_to_str(dmnbo[1]))
    if fnbo is not None:
        additional_sources.append(item_to_str(fnbo[0]))
        additional_targets.append(item_to_str(fnbo[1]))
    if dmmo is not None:
        additional_sources.append(item_to_str(dmmo[0]))
        additional_targets.append(item_to_str(dmmo[1]))
    if fmo is not None:
        additional_sources.append(item_to_str(fmo[0]))
        additional_targets.append(item_to_str(fmo[1]))

    input_name = item_to_str(input)
    parser_name = item_to_str(parser)

    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [input_name, *additional_sources],
        TARGET_ITEMS = [parser_name, *additional_targets],
    )

    def LoadNboResults_exec(**kw):
        log_fname = kw[input_name].access_element()
        if nbo_version == 3:
            parser_obj = NBO3LogParser(log_fname)
        elif nbo_version == 6:
            parser_obj = NBO6LogParser(log_fname)
        kw[parser_name].include_element(parser_obj)

        if dmnbo is not None:
            dmnbo_file = kw[dmnbo[0]].access_element()
            kw[dmnbo[1]].include_element(NboSymmMatrix(dmnbo_file, parser_obj.nbasis))

        if fnbo is not None:
            fnbo_file = kw[fnbo[0]].access_element()
            kw[fnbo[1]].include_element(NboSymmMatrix(fnbo_file, parser_obj.nbasis))

        if dmmo is not None:
            dmmo_file = kw[dmmo[0]].access_element()
            kw[dmmo[1]].include_element(NboSymmMatrix(dmmo_file, parser_obj.nbasis))

        if fmo is not None:
            fmo_file = kw[fmo[0]].access_element()
            kw[fmo[1]].include_element(NboSymmMatrix(fmo_file, parser_obj.nbasis))

        return ret.transform_finished()
    
    res.set_method({'exec': LoadNboResults_exec})
    return res


# Gaussian Cubegen
def gaussian_cubegen(name, fchk, output, orb_data, orb_index, nbo_log=None, nproc=1):
    additional_sources = []
    additional_sources_renamed = []
    additional_renames = []

    if nbo_log is not None:
        nbo_log_name = item_to_str(nbo_log)
        additional_sources.append('nbo_log')
        additional_sources_renamed.append(nbo_log_name)
        additional_renames.append(f'{nbo_log_name}->nbo_log')

    def init_cubegen(fchk, orb_data, calcdir, **kw):
        fchk_fname = fchk.access_element()
        fchk_tempfile = os.path.join(calcdir, 'start.fchk')
        shutil.copy2(fchk_fname, fchk_tempfile)
        
        orb_index = orb_data.access_element()['index']
        if nbo_log is not None:
            nbo_log_fname = kw['nbo_log'].access_element()
            final_orb_index = generate_reorder_map(nbo_log_fname, nbo_indices=[orb_index])[orb_index]
        else:
            final_orb_index = orb_index + 1

        return f"exec_gaussian_cubegen.sh {calcdir} {nproc} " \
            f"MO={final_orb_index} start.fchk result.cube 100"

    def finish_cubegen(output, calcdir, **kw):
        cube_temp_file = os.path.join(calcdir, 'result.cube')
        assert os.path.isfile(cube_temp_file), f"Cannot find file '{cube_temp_file}'"

        final_cube_name = output.get_path()
        shutil.copy2(cube_temp_file, final_cube_name)
        output.include_element(final_cube_name)

    template = nonblocking_subprocess(
        name='Cubegen_template',
        input=['fchk', 'orb_data', *additional_sources],
        output=['output'],
        command_prepare=init_cubegen,
        output_process=finish_cubegen,
        calcdir='calcdir',
        aware_keys=[orb_index],
        nproc=nproc
    )
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(fchk), item_to_str(orb_data), *additional_sources_renamed],
        TARGET_ITEMS = [item_to_str(output)],
        AWARE_KEYS = [orb_index]
    )
    res.extend(template, f"""\
        {item_to_str(fchk)}->fchk
        {item_to_str(output)}->output
        {item_to_str(orb_data)}->orb_data
        {' '.join(additional_renames)}
    """)
    return res

# MCubes calc
def run_mcubes():
    import os, sys, json
    sys.path.append(os.getcwd())

    from chemscripts.nbo.isosurfaces import generate_isosurface
    cube_fname = sys.argv[1]
    isurf_fnames = json.loads(sys.argv[2].replace("'", '"'))
    isovalue = float(sys.argv[3])
    generate_isosurface(cubename=cube_fname, meshfile_files=isurf_fnames, ival=isovalue)

def isurf_item_to_files(isurf_item, sign_key, type_key, method='get_path', **kwargs):
    from chemscripts.nbo.isosurfaces import ISURF_SIGNS, ISURF_TYPES
    isurf_fnames = {}
    for sign in ISURF_SIGNS:
        isurf_fnames[sign] = {}
        for itype in ISURF_TYPES:
            isurf_fnames[sign][itype] = getattr(isurf_item, method)(**{
                sign_key: sign,
                type_key: itype,
                **kwargs
            })
    return isurf_fnames

def mcubes_isosurface(name, input, pyfile, output, sign_key, type_key, isovalue=0.05):
    
    def argv_prepare(cube, isurf, **kw):
        cube_fname = cube.access_element()
        isurf_fnames = isurf_item_to_files(isurf, sign_key=sign_key, type_key=type_key)
        return (cube_fname, isurf_fnames, isovalue)
    
    def output_process(isurf, **kw):
        isurf_fnames = isurf_item_to_files(isurf, sign_key=sign_key, type_key=type_key)
        for sign, isurf_sign_data in isurf_fnames.items():
            for itype, filename in isurf_sign_data.items():
                assert os.path.isfile(filename), f"Cannot find file '{filename}'"
                isurf.include_element(filename, **{
                    sign_key: sign,
                    type_key: itype,
                })
                
    template = pyfunction_subprocess(
        name='MCubestemplate',
        input='cube',
        output='isurf',
        pyfunction=run_mcubes,
        argv_prepare=argv_prepare,
        output_process=output_process,
        pyfile=pyfile,
        nproc=1
    )
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output), item_to_str(pyfile)]
    )
    res.extend(template, f"""\
        {item_to_str(input)}->cube
        {item_to_str(pyfile)}->pyfile
        {item_to_str(output)}->isurf
    """)
    return res


#
# BLENDER
#
DEFAULT_STYLES = {
    'red-blue': {
        'plus': {
            'color': "#DF4A4A",
        },
        'minus': {
            'color': "#64A7E7",
        }
    },
    'green-yellow': {
        'plus': {
            'color': "#5CCA59",
            'alpha': 0.2,
            'factor': 0.6,
        },
        'minus': {
            'color': "#FFA624",
            'alpha': 0.2,
            'factor': 0.6,
        }
    },
}

def blender_main_script():
    import bpy

    import sys
    module_dir = INSERT_HERE
    if module_dir not in sys.path:
        sys.path.insert(0, module_dir)

    scene_spec = INSERT_HERE
    STYLES = INSERT_HERE

    if scene_spec is not None:
        import chemscripts.blender as msblend
        # import chemscripts.blender.draw_objects.linal as mslinal

        DEFAULTS = {
            'orbitals': {},
            'molecules': {},
        }
        MOLECULE_KEYS = ['xyz', 'symbols', 'bonds', 'bondtypes', 'smallatoms']
        MOLECULE_DEFAULTS = {
            'smallatoms': []
        }
        
        # Primary checks
        assert not ('molecules' in scene_spec and any(key in scene_spec for key in MOLECULE_KEYS)), \
            f"Molecule can be specified via 'molecules' key or directly via '{MOLECULE_KEYS}', but not both"

        for key, value in DEFAULTS.items():
            if key not in scene_spec:
                scene_spec[key] = value
        
        if any(key in scene_spec for key in MOLECULE_KEYS):
            default_molname = 'mol'
            scene_spec['molecules'][default_molname] = {}
            for key in MOLECULE_KEYS:
                if key not in scene_spec:
                    continue
                scene_spec['molecules'][default_molname][key] = scene_spec[key]
                del scene_spec[key]
        for mol_name, mol_data in scene_spec['molecules'].items():
            for key, default_value in MOLECULE_DEFAULTS.items():
                if key not in mol_data:
                    mol_data[key] = default_value
            assert all(key in mol_data for key in MOLECULE_KEYS), \
                f"Molecule '{mol_name}' is partially specified: {mol_data}"

        msblend.cleanup(protect_collections=[
            mol_name
            for mol_name in scene_spec['molecules'].keys()
        ])
        for mol_name, mol_data in scene_spec['molecules'].items():
            m = msblend.Molecule(name=mol_name, small_atoms=mol_data['smallatoms'])
            m.from_dict(mol_data)
            bonds_obj = m.draw_bonds(caps=False, radius=0.1)
            atoms_obj = m.draw_atoms(scale=0.35)
            mol_collection = bpy.data.collections.new(mol_name)
            bpy.context.scene.collection.children.link(mol_collection)
            bpy.data.collections['Collection'].children.link(mol_collection)

            if bonds_obj.name not in bpy.data.collections[mol_name].objects:
                bpy.data.collections[mol_name].objects.link(bonds_obj)
            if atoms_obj.name not in bpy.data.collections[mol_name].objects:
                bpy.data.collections[mol_name].objects.link(atoms_obj)

        for name, style_data in STYLES.items():
            style_data['name'] = name

        for name, isurf_data in scene_spec['orbitals'].items():
            import chemscripts.blender.draw_objects.orbitals as msorbital
            msorbital.plot_nbo(name, isurf_data['files'], style=STYLES[isurf_data['style']], reverse=False)

    scene = bpy.context.scene

    BLEND_FINAL = INSERT_HERE
    if BLEND_FINAL is not None:
        bpy.ops.wm.save_as_mainfile(filepath=BLEND_FINAL)

    PNG_FINAL = INSERT_HERE
    NPROC = INSERT_HERE
    if PNG_FINAL is not None:
        scene.render.image_settings.file_format = 'PNG'
        scene.render.filepath = PNG_FINAL
        
        try:
            bpy.context.preferences.addons[
                "cycles"
            ].preferences.compute_device_type = "CUDA"
        except:
            pass

        if NPROC is not None:
            bpy.context.scene.render.threads = NPROC
        else:
            bpy.context.scene.render.threads = 1

        bpy.context.scene.render.threads_mode = 'FIXED'
        bpy.ops.render.render(write_still=1)

def blender_render(name, blend, png, nproc=1, **kw):
    return blender_for_scenes(
        name=name,
        scene_template=blend,
        png=png,
        nproc=nproc,
        **kw
    )

def blender_for_scenes(name,
        scene_template,
        scene_spec=None,
        blend=None,
        script=None,
        png=None,
        aware_keys=[],
        nproc=1,
        styles=None
    ):
    assert utils.check_availability('exec_blender.sh')
    assert blend is not None or png is not None

    additional_sources = []
    additional_sources_renamed = []
    additional_targets = []
    additional_targets_renamed = []
    additional_renames = []

    template_kwargs = {}

    if scene_spec is not None:
        scene_spec_name = item_to_str(scene_spec)
        additional_sources.append('scene_spec')
        additional_sources_renamed.append(scene_spec_name)
        additional_renames.append(f'{scene_spec_name}->scene_spec')

    if script is not None:
        script_name = item_to_str(script)
        additional_targets_renamed.append(script_name)
        additional_renames.append(f'{script_name}->script')
        template_kwargs['pyfile'] = 'script'

    if blend is not None:
        blend_name = item_to_str(blend)
        additional_targets.append('blend')
        additional_targets_renamed.append(blend_name)
        additional_renames.append(f'{blend_name}->blend')

    if png is not None:
        png_name = item_to_str(png)
        additional_targets.append('png')
        additional_targets_renamed.append(png_name)
        additional_renames.append(f'{png_name}->png')

    def blender_preparation(file, # This is the script filename
            scene_template, calcdir, **kw
        ):
        scene_template_path = scene_template.access_element()
        temp_template = os.path.join(calcdir, 'template.blend')
        shutil.copy2(scene_template_path, temp_template)

        script_subs = {
            'module_dir': os.getcwd(),
            'NPROC': nproc,
        }
        if styles is not None:
            script_subs['STYLES'] = styles
        else:
            script_subs['STYLES'] = DEFAULT_STYLES

        if scene_spec is not None:
            script_subs['scene_spec'] = kw['scene_spec'].access_element()
        else:
            script_subs['scene_spec'] = None

        if png is not None:
            script_subs['PNG_FINAL'] = 'done.png'
        else:
            script_subs['PNG_FINAL'] = None

        if blend is not None:
            script_subs['BLEND_FINAL'] = 'done.blend'
        else:
            script_subs['BLEND_FINAL'] = None

        script_code = function_to_script(blender_main_script, subs=script_subs)
        with open(file, 'w') as f:
            f.write(script_code)

        return f"exec_blender.sh {calcdir} template.blend -b -P {ntpath.basename(file)}"

    def output_process(calcdir, **kw):
        if blend is not None:
            temp_blend = os.path.join(calcdir, 'done.blend')
            assert os.path.isfile(temp_blend), f"Cannot find file '{temp_blend}'"
            blend_final_path = kw['blend'].get_path()
            shutil.copy2(temp_blend, blend_final_path)
            kw['blend'].include_element(blend_final_path)
        
        if png is not None:
            temp_png = os.path.join(calcdir, 'done.png')
            assert os.path.isfile(temp_png), f"Cannot find file '{temp_png}'"
            png_final_path = kw['png'].get_path()
            shutil.copy2(temp_png, png_final_path)
            kw['png'].include_element(png_final_path)

    template = pyfunction_subprocess(
        name='Blendertemplate',
        input=['scene_template', *additional_sources],
        output=additional_targets,
        custom_command=blender_preparation,
        output_process=output_process,
        nproc=nproc,
        calcdir='calcdir',
        aware_keys=aware_keys,
        **template_kwargs
    )

    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(scene_template), *additional_sources_renamed],
        TARGET_ITEMS = [*additional_targets_renamed],
        AWARE_KEYS = aware_keys
    )
    res.extend(template, f"""\
        {item_to_str(scene_template)}->scene_template
        {' '.join(additional_renames)}
    """)
    return res

def png_postprocess(name, input, output, settings, merged_keys=[]):
    template = Transform(
        NAME = 'PngProcesstemplate',
        SOURCE_ITEMS = ['startpng'],
        TARGET_ITEMS = ['donepng'],
        MERGED_KEYS = merged_keys,
    )

    DEFAULTS = {
        'fill_transparency': True,
        'brightness': None,
        'trim_white': True,
        'white_to_transparent': False,
        'shrink_factor': None,
    }
    final_settings = {
        key: settings[key] if key in settings else default_value
        for key, default_value in DEFAULTS.items()
    }

    def PngProcess_exec(startpng, donepng):
        from PIL import Image, ImageEnhance
        from chemscripts import imageutils

        assert len(merged_keys) == 0 or final_settings['trim_white']

        images = []
        for startpng_fname, keys in startpng:
            main_image = Image.open(startpng_fname)
            main_image.load()

            # Fills the transparency
            if final_settings['fill_transparency']:
                bg = Image.new("RGB", main_image.size, (255, 255, 255))
                bg.paste(main_image, mask=main_image.split()[3])
                main_image = bg

            # Increases brightness
            if final_settings['brightness'] is not None:
                enhancer = ImageEnhance.Brightness(main_image)
                main_image = enhancer.enhance(final_settings['brightness'])
            
            if final_settings['white_to_transparent']:
                main_image = main_image.convert("RGBA")
                data = main_image.getdata()
                new_data = []
                for item in data:
                    if item[:3] == (255, 255, 255):
                        new_data.append((255, 255, 255, 0))
                    else:
                        new_data.append(item)
                main_image.putdata(new_data)
            
            if final_settings['shrink_factor'] is not None:
                factor = final_settings['shrink_factor']
                main_image = main_image.resize((int(main_image.width / factor), int(main_image.height / factor)))
            images.append({
                'obj': main_image,
                'keys': keys
            })

        # Trims white space
        if final_settings['trim_white']:
            trimbox = imageutils.TrimmingBox()
            for elem in images:
                trimbox.extend(elem['obj'])
            for elem in images:
                elem['obj'] = elem['obj'].crop(trimbox.points)


        for elem in images:
            keys = elem['keys']
            donepng_fname = donepng.get_path(**keys)
            elem['obj'].save(donepng_fname, 'PNG', quality=100)
            donepng.include_element(donepng_fname, **keys)

        return ret.transform_finished()

    template.set_method({'exec': PngProcess_exec})
    
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
        MERGED_KEYS = merged_keys,
    )
    res.extend(template, f"""\
        {item_to_str(input)}->startpng
        {item_to_str(output)}->donepng
    """)
    return res

#
# FILE OPERATIONS
#

# Copy single files
def copy_file(name, input, output, condition=None):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )

    def CopyFile_exec(start, finish):
        start_fname = start.access_element()
        assert os.path.isfile(start_fname), f"Source file '{start_fname}' does not exist"

        if condition is None or condition(start_fname):
            finish_fname = finish.get_path()
            shutil.copy2(start_fname, finish_fname)
            finish.include_element(finish_fname)
        return ret.transform_finished()

    CopyFile_template = Transform(
        NAME = 'CopyFiletemplate',
        SOURCE_ITEMS = ['start'],
        TARGET_ITEMS = ['finish'],
    )
    CopyFile_template.set_method({'exec': CopyFile_exec})

    res.extend(CopyFile_template, f"""\
        {item_to_str(input)}->start
        {item_to_str(output)}->finish
    """)
    return res


# Extract archive
def extract_archive(name, input, output, aware_keys=[], filename=None):
    ExtractArchive_template = Transform(
        NAME = 'ExtractArchivetemplate',
        SOURCE_ITEMS = ['archive'],
        TARGET_ITEMS = ['extracted_files'],
        AWARE_KEYS = aware_keys,
    )

    def ExtractArchive_exec(archive, extracted_files, **kw):
        start_fname = archive.access_element()
        assert os.path.isfile(start_fname), f"Source file '{start_fname}' does not exist"
        
        diritem = extracted_files.containing_dir()
        extract_dir = diritem.get_path()
        if diritem.restriction_id is not None:
            diritem.retire()

        # To create containing dir
        extracted_files.get_path(**{key: '' for key in extracted_files.public_keys})

        with tarfile.open(start_fname, 'r:gz') as tar:
            for member in tar.getmembers():
                if filename is not None:
                    member.name = filename(
                        original_name=member.name,
                        **kw
                    )
                tar.extract(member, path=extract_dir)
                extracted_files.include_element(os.path.join(extract_dir, member.name))
        return ret.transform_finished()

    ExtractArchive_template.set_method({'exec': ExtractArchive_exec})

    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
        AWARE_KEYS = aware_keys,
    )

    res.extend(ExtractArchive_template, f"""\
        {item_to_str(input)}->archive
        {item_to_str(output)}->extracted_files
    """)
    return res

def load_json(name, input, output):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
    )

    def ReadJson_exec(input, output):
        input_fname = input.access_element()
        with open(input_fname, 'r') as f:
            data = json.load(f)
        
        output_keys_set = set(output.public_keys)
        if len(output_keys_set) == 0:
            output.include_element(data)
        else:
            for elem in data:
                keys = elem['keys']
                assert set(keys.keys()) == output_keys_set, \
                    f"Generated keys mismatch: '{keys}' vs. '{output_keys_set}'"
                output.include_element(elem['obj'], **keys)

        return ret.transform_finished()

    ReadJson_template = Transform(
        NAME = 'ReadJsontemplate',
        SOURCE_ITEMS = ['input'],
        TARGET_ITEMS = ['output'],
    )
    ReadJson_template.set_method({'exec': ReadJson_exec})

    res.extend(ReadJson_template, f"""\
        {item_to_str(input)}->input
        {item_to_str(output)}->output
    """)
    return res


def dump_json(name, input, output, merged_keys=[]):
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
        MERGED_KEYS = merged_keys,
    )

    merged_keys_set = set(merged_keys)

    def SaveJson_exec(input, output):
        if len(merged_keys) == 0:
            input_obj = input.access_element()
            save_obj = input_obj
        else:
            for obj, keys in input:
                assert set(keys.keys()) == merged_keys_set, \
                    f"Merged keys mismatch: '{keys}' vs. '{merged_keys_set}'"
            save_obj = [
                {
                    'obj': obj,
                    'keys': keys
                }
                for obj, keys in input
            ]

        output_fname = output.get_path()
        with open(output_fname, 'w') as f:
            json.dump(save_obj, f)
        output.include_element(output_fname)
        return ret.transform_finished()

    SaveJson_template = Transform(
        NAME = 'SaveJsontemplate',
        SOURCE_ITEMS = ['input'],
        TARGET_ITEMS = ['output'],
        MERGED_KEYS = merged_keys,
    )
    SaveJson_template.set_method({'exec': SaveJson_exec})

    res.extend(SaveJson_template, f"""\
        {item_to_str(input)}->input
        {item_to_str(output)}->output
    """)
    return res

def save_pandas(df, csvname):
    df.to_csv(csvname, index=False)
    return csvname

def pd_to_csv(name, input, output):
    return map(
        name=name,
        input=input,
        output=output,
        mapping=lambda **kw: save_pandas(df=kw[input], csvname=kw[output])
    )

def load_pandas(csvname):
    return pd.read_csv(csvname)

def pd_from_csv(name, input, output):
    return map(
        name=name,
        input=input,
        output=output,
        mapping=lambda **kw: load_pandas(csvname=kw[output])
    )

#
# MOVIES
#

def frames_to_mp4(name, input, output, ordering_key=None, aware_keys=[], merged_keys=[]):
    assert len(merged_keys) > 0, \
        f"Transition frames->mp4 implied the presence of at least one merged key but got none"
    
    template = Transform(
        NAME = name,
        SOURCE_ITEMS = ['frames'],
        TARGET_ITEMS = ['mp4'],
        AWARE_KEYS = aware_keys,
        MERGED_KEYS = merged_keys
    )

    def GetMP4_exec(frames, mp4, thread_manager, **kw):
        frames_data = [
            {
                'file': elem,
                'keys': keys,
            }
            for elem, keys in frames
        ]
        if ordering_key is not None:
            frames_data.sort(key=ordering_key)
        else:
            png_keys = frames.public_keys
            assert len(png_keys) == 1, f"Unable to sort automatically when have multiple keys '{png_keys}'. Specify 'ordering_key'"
            png_key = png_keys[0]
            frames_data.sort(key=lambda obj: obj['keys'][png_key])
        
        id, used_calcdir = thread_manager.prepare_new_job()
        number_of_elements = len(frames_data)
        number_of_digits = len(str(number_of_elements)) + 1
        frame_mask = f'frame_%{number_of_digits}d.png'
        frame_path_mask = os.path.join(used_calcdir, frame_mask)
        for i, item in enumerate(frames_data):
            raw_name = frame_path_mask % i
            temp_name = os.path.join(os.path.dirname(raw_name), ntpath.basename(raw_name).replace(' ', '0'))
            shutil.copy2(item['file'], temp_name)
        
        main_dir = os.getcwd()
        os.chdir(used_calcdir)
        command = f'ffmpeg -y -framerate 10 -i {frame_mask} -c:v libx264 -crf 17 -vf "format=yuv420p,pad=ceil(iw/2)*2:ceil(ih/2)*2" result.mp4'
        os.system(command)
        temp_result_file = os.path.join(used_calcdir, 'result.mp4')

        final_file = mp4.get_path()
        assert os.path.isfile(temp_result_file), f"Result file '{temp_result_file}' is not found"
        shutil.copy2(temp_result_file, final_file)
        mp4.include_element(final_file)
        
        os.chdir(main_dir)
        thread_manager.finalize_task(id, state='registered')
        return ret.transform_finished()

    template.set_method({'exec': GetMP4_exec})
    res = Transform(
        NAME = name,
        SOURCE_ITEMS = [item_to_str(input)],
        TARGET_ITEMS = [item_to_str(output)],
        AWARE_KEYS = aware_keys,
        MERGED_KEYS = merged_keys
    )
    res.extend(template, f"""\
        {item_to_str(input)}->frames
        {item_to_str(output)}->mp4
    """)
    return res


def frames_to_gif(name, input, output, aware_keys=[], merged_keys=[]):
    assert len(merged_keys) > 0, \
        f"Transition frames->mp4 implied the presence of at least one merged key but got none"



import os

from .decorators import register_dataitem
from .abstract_path import PathItem

from chemscripts.utils import str_to_type


@register_dataitem('dirs')
class DirectoryItem(PathItem):
    def __init__(self,
            directories: list = None,
            key_requests: dict = {},
            skip_autoindexing: bool = False,
            **kwargs
        ):
        """

        Args:
            mask (str): filename mask of the form like '{keyA}_{keyB}.ext'.
            wd (str): directory where all FileItem files are stored.
            files (List[str], optional): restrictions to load only specific files
            key_requests (dict, optional): key restrictions for files loading.
            Cannot be provided together with `files` kwarg.
        """
        super().__init__(**kwargs)

        # Will be used to finish initialization in the `checkin` method
        self.checkin_kwargs = {
            'directories': directories,
            'key_requests': key_requests,
            'skip_autoindexing': skip_autoindexing,
        }

    def checkin(self) -> None:
        """Since this part of initialization requires interaction with `self.storage`
        it is executed a bit after the __init__ call
        """
        super().checkin()
        
        if not self.checkin_kwargs['skip_autoindexing']:
            self._index_elements(
                key_requests=self.checkin_kwargs['key_requests'],
                required_elements=self.checkin_kwargs['directories'],
            )

    def ensure_directory(self, **keys):
        # This will perform some basic checks of **keys
        self.get_path(**keys)

        # Create directory if it doesn't exist already
        self.storage.verify_directory(diritem=self, **self._restriction_request(key_restrictions=keys))
        
    def cleanup(self) -> None:
        for dirname, keys in self:
            assert os.path.isdir(dirname), f'{dirname} not a directory'
            if len(os.listdir(dirname)) == 0:
                os.rmdir(dirname)
        
        # This will remove the table elements corresponding to non-existing paths
        super().cleanup()
    
    def deallocate_directory(self, dirname):
        assert dirname in self
    
    def validate_path(self, path: str) -> bool:
        return os.path.isdir(path)





class DotDict:
    def __init__(self, dictionary):
        self.__dict__.update(dictionary)

def get_logger_shortcuts(logger):
    if logger is not None:
        return DotDict({
            'debug': lambda message: logger.debug(message),
            'info': lambda message: logger.info(message),
            'warning': lambda message: logger.warning(message),
            'error': lambda message: logger.error(message),
        })
    else:
        return DotDict({
            'debug': lambda message: None,
            'info': lambda message: None,
            'warning': lambda message: None,
            'error': lambda message: None,
        })



REGISTERED_DATAITEM_TYPES = {}

CHILD_PROPERTIES = {
    'object': {
        'is_file': False,
        'is_object': True,
        'is_path': False
    },
    'file': {
        'is_file': True,
        'is_object': False,
    },
    'path': {
        'is_path': True,
        'is_object': False,
    }
}

def register_dataitem(class_alias: str):
    """TODO Docstring
    """
    def decorator(class_object):
        REGISTERED_DATAITEM_TYPES[class_alias] = class_object

        # Add @property methods to class, like is_file, is_object and
        # everything else that was specified in CHILD_PROPERTIES
        if class_alias in CHILD_PROPERTIES:
            for prop_name, prop_value in CHILD_PROPERTIES[class_alias].items():
                def create_property_method(value):
                    @property
                    def getter(self):
                        return value
                    return getter

                prop_getter = create_property_method(prop_value)
                setattr(class_object, prop_name, prop_getter)

        return class_object
    return decorator



import threading
import textwrap3
import time
import os
import traceback
import queue
from io import StringIO

from prompt_toolkit import Application
from prompt_toolkit.layout.containers import VSplit, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.key_binding import KeyBindings

from chemscripts.mylogging import createLogger


class CustomStringIO(StringIO):
    def __init__(self, *args, func=None, cleanup: bool=False, **kwargs):
        self.func = func
        self.cleanup = cleanup
        super().__init__(*args, **kwargs)

    def write(self, *args, **kwargs):
        if self.cleanup:
            self.truncate(0)
            self.seek(0)
        super().write(*args, **kwargs)
        self.func(self.getvalue())


kb = KeyBindings()
@kb.add('c-c')
def exit_(event):
    event.app.exit()

left_panel = FormattedTextControl("")
right_panel = FormattedTextControl("")

left_window = Window(left_panel)
right_window = Window(right_panel)
root_container = VSplit([
    left_window,
    Window(width=1, char='|'),
    right_window,
])

layout = Layout(root_container)

app = Application(layout=layout, key_bindings=kb, refresh_interval=0.05, full_screen=True)

def run_app():
    app.run()


def get_terminal_size():
    try:
        # Get terminal size using the 'stty' command (Unix-based systems)
        rows, columns = os.popen('stty size', 'r').read().split()
        return int(rows), int(columns)
    except:
        pass

    try:
        # Get terminal size using the 'ioctl' system call (Windows)
        import struct
        from fcntl import ioctl
        from termios import TIOCGWINSZ

        fd = os.open(os.ctermid(), os.O_RDONLY)
        size = struct.unpack('hh', ioctl(fd, TIOCGWINSZ, '1234'))
        os.close(fd)
        return size[0], size[1]
    except OSError:
        pass

    # Fallback in case both methods fail
    return 30, 120

NUM_LINES = int(get_terminal_size()[0])
WIDTH = int((get_terminal_size()[1] - 2) / 2)

def wrap_text(t: str) -> str:
    x = '\n'.join([
        textwrap3.fill(line, width=WIDTH)
        for line in t.splitlines()
    ])
    return '\n'.join(x.splitlines())


def custom_update(panel_name, show_part):
    if panel_name == 'left':
        panel = left_panel
    elif panel_name == 'right':
        panel = right_panel
    else:
        raise RuntimeError(f"Unknown panel '{panel_name}'")

    if show_part == 'top':
        cropping_function = lambda text: '\n'.join(text.splitlines()[:NUM_LINES])
    elif show_part == 'bottom':
        cropping_function = lambda text: '\n'.join(text.splitlines()[-NUM_LINES:])
    else:
        raise RuntimeError(f"Unknown cropping setting '{show_part}'")
    
    def update_function(newtext: str) -> None:
        panel.text = cropping_function(wrap_text(cropping_function(newtext)))
    return update_function


def execute_with_ui(main_function):
    logger_stream = CustomStringIO(func=custom_update(panel_name='right', show_part='bottom'))
    main_logger = createLogger("Main", logger_stream)
    stack_stream = CustomStringIO(func=custom_update(panel_name='left', show_part='top'), cleanup=True)

    app_thread = threading.Thread(target=run_app)
    app_thread.start()
 
    exception_queue = queue.Queue()
    def run_main():
        try:
            main_function(main_logger, stack_stream)
        except:
            exception_queue.put(traceback.format_exc())

    work_thread = threading.Thread(target=run_main)
    work_thread.start()
    work_thread.join()

    time.sleep(0.3)
    app.exit()
    app_thread.join()

    print(logger_stream.getvalue())

    if not exception_queue.empty():
        exception = exception_queue.get()
        print(f"Worker thread raised an exception:\n{exception}")
